from optparse import OptionGroup
from logging import (getLogger, StreamHandler,
                     DEBUG, INFO, WARNING, ERROR)
from sys import exit
from ptrace import PtraceError
from logging import error
from ptrace.tools import locateProgram
from ptrace.debugger import ProcessExit, DebuggerError
from errno import EPERM
from ptrace.debugger.child import createChild


class Application(object):

    def __init__(self):
        pass

    def _setupLog(self, fd):
        logger = getLogger()
        handler = StreamHandler(fd)
        logger.addHandler(handler)
        if self.options.debug:
            level = DEBUG
        elif self.options.verbose:
            level = INFO
        elif self.options.quiet:
            level = ERROR
        else:
            level = WARNING
        logger.setLevel(level)

    def processOptions(self):
        if self.program:
            self.program[0] = locateProgram(self.program[0])

    def createLogOptions(self, parser):
        log = OptionGroup(parser, "Logging")
        log.add_option("--quiet", "-q", help="Be quiet (set log level to ERROR)",
                       action="store_true", default=False)
        log.add_option("--verbose", "-v", help="Debug mode (set log level to INFO)",
                       action="store_true", default=False)
        log.add_option("--debug", help="Debug mode (set log level to DEBUG)",
                       action="store_true", default=False)
        parser.add_option_group(log)

    def createChild(self, arguments, env=None):
        return createChild(arguments, self.options.no_stdout, env)

    def setupDebugger(self):
        # Set ptrace options
        if self.options.fork:
            try:
                self.debugger.traceFork()
            except DebuggerError:
                error("--fork option is not supported by your OS, sorry!")
                exit(1)
        if self.options.trace_exec:
            self.debugger.traceExec()
        if self.options.trace_clone:
            self.debugger.traceClone()

    def createProcess(self):
        if self.options.pid:
            pid = self.options.pid
            is_attached = False
            error("Attach process %s" % pid)
        else:
            pid = self.createChild(self.program)
            is_attached = True
        try:
            return self.debugger.addProcess(pid, is_attached=is_attached)
        except (ProcessExit, PtraceError) as err:
            if isinstance(err, PtraceError) \
                    and err.errno == EPERM:
                error(
                    "ERROR: You are not allowed to trace process %s (permission denied or process already traced)" % pid)
            else:
                error("ERROR: Process can no be attached! %s" % err)
        return None

    def createCommonOptions(self, parser):
        parser.add_option("--pid", "-p", help="Attach running process specified by its identifier",
                          type="int", default=None)
        parser.add_option("--fork", "-f", help="Trace fork() event and child processes",
                          action="store_true", default=False)
        parser.add_option("--trace-exec", help="Trace execve() event",
                          action="store_true", default=False)
        parser.add_option("--trace-clone", help="Trace clone() event",
                          action="store_true", default=False)
        parser.add_option("--no-stdout", help="Use /dev/null as stdout/stderr, or close stdout and stderr if /dev/null doesn't exist",
                          action="store_true", default=False)



# From Linux kernel source code
#    include/linux/syscalls.h
#    arch/i386/kernel/syscall_table.S
#    arch/um/include/sysdep-i386/syscalls.h
#    arch/um/sys-i386/sys_call_table.S

ALIASES = {
    "brk": ("break",),
    "fadvise64": ("posix_fadvise",),
    "fstatat64": ("fstatat",),
    "getcwd": ("__getcwd",),
    "mmap_pgoff": ("mmap", "mmap2",),
    "pread64": ("pread",),
    "prlimit64": ("prlimit",),
    "pselect6": ("pselect",),
    "pwrite64": ("pwrite",),
}

# Name of arguments containing a filename or a path
FILENAME_ARGUMENTS = set(
    ("filename", "pathname", "path", "oldname", "newname", "old", "new"))

SYSCALL_PROTOTYPES = {
    "accept": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int *", "addrlen"),
    )),
    "accept4": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int *", "addrlen"),
        ("int", "flags"),
    )),
    "access": ("long", (
        ("const char *", "filename"),
        ("int", "mode"),
    )),
    "acct": ("long", (
        ("const char *", "name"),
    )),
    "add_key": ("long", (
        ("const char *", "_type"),
        ("const char *", "_description"),
        ("const void *", "_payload"),
        ("size_t", "plen"),
        ("key_serial_t", "destringid"),
    )),
    "adjtimex": ("long", (
        ("struct timex *", "txc_p"),
    )),
    "alarm": ("long", (
        ("unsigned int", "seconds"),
    )),
    "bdflush": ("long", (
        ("int", "func"),
        ("long", "data"),
    )),
    "bind": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int", "addrlen"),
    )),
    "bpf": ("long", (
        ("int", "cmd"),
        ("union bpf_attr *", "attr"),
        ("unsigned int", "size"),
    )),
    "brk": ("long", (
        ("unsigned long", "brk"),
    )),
    "capget": ("long", (
        ("cap_user_header_t", "header"),
        ("cap_user_data_t", "dataptr"),
    )),
    "capset": ("long", (
        ("cap_user_header_t", "header"),
        ("const cap_user_data_t", "data"),
    )),
    "chdir": ("long", (
        ("const char *", "filename"),
    )),
    "chmod": ("long", (
        ("const char *", "filename"),
        ("umode_t", "mode"),
    )),
    "chown": ("long", (
        ("const char *", "filename"),
        ("uid_t", "user"),
        ("gid_t", "group"),
    )),
    "chown16": ("long", (
        ("const char *", "filename"),
        ("old_uid_t", "user"),
        ("old_gid_t", "group"),
    )),
    "chroot": ("long", (
        ("const char *", "filename"),
    )),
    "clock_adjtime": ("long", (
        ("clockid_t", "which_clock"),
        ("struct timex *", "tx"),
    )),
    "clock_getres": ("long", (
        ("clockid_t", "which_clock"),
        ("struct timespec *", "tp"),
    )),
    "clock_gettime": ("long", (
        ("clockid_t", "which_clock"),
        ("struct timespec *", "tp"),
    )),
    "clock_nanosleep": ("long", (
        ("clockid_t", "which_clock"),
        ("int", "flags"),
        ("const struct timespec *", "rqtp"),
        ("struct timespec *", "rmtp"),
    )),
    "clock_settime": ("long", (
        ("clockid_t", "which_clock"),
        ("const struct timespec *", "tp"),
    )),
    "clone": ("long", (
        ("unsigned long", "flags"),
        ("unsigned long", "child_stack"),
        ("int *", "ptid"),
        ("int *", "ctid"),
        ("unsigned long", "regs"),
    )),
    "close": ("long", (
        ("unsigned int", "fd"),
    )),
    "connect": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int", "addrlen"),
    )),
    "copy_file_range": ("long", (
        ("int", "fd_in"),
        ("loff_t *", "off_in"),
        ("int", "fd_out"),
        ("loff_t *", "off_out"),
        ("size_t", "len"),
        ("unsigned int", "flags"),
    )),
    "creat": ("long", (
        ("const char *", "pathname"),
        ("umode_t", "mode"),
    )),
    "delete_module": ("long", (
        ("const char *", "name_user"),
        ("unsigned int", "flags"),
    )),
    "dup": ("long", (
        ("unsigned int", "fildes"),
    )),
    "dup2": ("long", (
        ("unsigned int", "oldfd"),
        ("unsigned int", "newfd"),
    )),
    "dup3": ("long", (
        ("unsigned int", "oldfd"),
        ("unsigned int", "newfd"),
        ("int", "flags"),
    )),
    "epoll_create": ("long", (
        ("int", "size"),
    )),
    "epoll_create1": ("long", (
        ("int", "flags"),
    )),
    "epoll_ctl": ("long", (
        ("int", "epfd"),
        ("int", "op"),
        ("int", "fd"),
        ("struct epoll_event *", "event"),
    )),
    "epoll_pwait": ("long", (
        ("int", "epfd"),
        ("struct epoll_event *", "events"),
        ("int", "maxevents"),
        ("int", "timeout"),
        ("const sigset_t *", "sigmask"),
        ("size_t", "sigsetsize"),
    )),
    "epoll_wait": ("long", (
        ("int", "epfd"),
        ("struct epoll_event *", "events"),
        ("int", "maxevents"),
        ("int", "timeout"),
    )),
    "eventfd": ("long", (
        ("unsigned int", "count"),
    )),
    "eventfd2": ("long", (
        ("unsigned int", "count"),
        ("int", "flags"),
    )),
    "execve": ("long", (
        ("const char *", "filename"),
        ("const char *const *", "argv"),
        ("const char *const *", "envp"),
    )),
    "execveat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("const char *const *", "argv"),
        ("const char *const *", "envp"),
        ("int", "flags"),
    )),
    "exit": ("long", (
        ("int", "error_code"),
    )),
    "exit_group": ("long", (
        ("int", "error_code"),
    )),
    "faccessat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("int", "mode"),
    )),
    "fadvise64": ("long", (
        ("int", "fd"),
        ("loff_t", "offset"),
        ("size_t", "len"),
        ("int", "advice"),
    )),
    "fadvise64_64": ("long", (
        ("int", "fd"),
        ("loff_t", "offset"),
        ("loff_t", "len"),
        ("int", "advice"),
    )),
    "fallocate": ("long", (
        ("int", "fd"),
        ("int", "mode"),
        ("loff_t", "offset"),
        ("loff_t", "len"),
    )),
    "fanotify_init": ("long", (
        ("unsigned int", "flags"),
        ("unsigned int", "event_f_flags"),
    )),
    "fanotify_mark": ("long", (
        ("int", "fanotify_fd"),
        ("unsigned int", "flags"),
        ("u64", "mask"),
        ("int", "fd"),
        ("const char *", "pathname"),
    )),
    "fchdir": ("long", (
        ("unsigned int", "fd"),
    )),
    "fchmod": ("long", (
        ("unsigned int", "fd"),
        ("umode_t", "mode"),
    )),
    "fchmodat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("umode_t", "mode"),
    )),
    "fchown": ("long", (
        ("unsigned int", "fd"),
        ("uid_t", "user"),
        ("gid_t", "group"),
    )),
    "fchown16": ("long", (
        ("unsigned int", "fd"),
        ("old_uid_t", "user"),
        ("old_gid_t", "group"),
    )),
    "fchownat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("uid_t", "user"),
        ("gid_t", "group"),
        ("int", "flag"),
    )),
    "fcntl": ("long", (
        ("unsigned int", "fd"),
        ("unsigned int", "cmd"),
        ("unsigned long", "arg"),
    )),
    "fcntl64": ("long", (
        ("unsigned int", "fd"),
        ("unsigned int", "cmd"),
        ("unsigned long", "arg"),
    )),
    "fdatasync": ("long", (
        ("unsigned int", "fd"),
    )),
    "fgetxattr": ("long", (
        ("int", "fd"),
        ("const char *", "name"),
        ("void *", "value"),
        ("size_t", "size"),
    )),
    "finit_module": ("long", (
        ("int", "fd"),
        ("const char *", "uargs"),
        ("int", "flags"),
    )),
    "flistxattr": ("long", (
        ("int", "fd"),
        ("char *", "list"),
        ("size_t", "size"),
    )),
    "flock": ("long", (
        ("unsigned int", "fd"),
        ("unsigned int", "cmd"),
    )),
    "fork": ("long", (
    )),
    "fremovexattr": ("long", (
        ("int", "fd"),
        ("const char *", "name"),
    )),
    "fsetxattr": ("long", (
        ("int", "fd"),
        ("const char *", "name"),
        ("const void *", "value"),
        ("size_t", "size"),
        ("int", "flags"),
    )),
    "fstat": ("long", (
        ("unsigned int", "fd"),
        ("struct __old_kernel_stat *", "statbuf"),
    )),
    "fstat64": ("long", (
        ("unsigned long", "fd"),
        ("struct stat64 *", "statbuf"),
    )),
    "fstatat64": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("struct stat64 *", "statbuf"),
        ("int", "flag"),
    )),
    "fstatfs": ("long", (
        ("unsigned int", "fd"),
        ("struct statfs *", "buf"),
    )),
    "fstatfs64": ("long", (
        ("unsigned int", "fd"),
        ("size_t", "sz"),
        ("struct statfs64 *", "buf"),
    )),
    "fsync": ("long", (
        ("unsigned int", "fd"),
    )),
    "ftruncate": ("long", (
        ("unsigned int", "fd"),
        ("unsigned long", "length"),
    )),
    "ftruncate64": ("long", (
        ("unsigned int", "fd"),
        ("loff_t", "length"),
    )),
    "futex": ("long", (
        ("u32 *", "uaddr"),
        ("int", "op"),
        ("u32", "val"),
        ("struct timespec *", "utime"),
        ("u32 *", "uaddr2"),
        ("u32", "val3"),
    )),
    "futimesat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("struct timeval *", "utimes"),
    )),
    "get_mempolicy": ("long", (
        ("int *", "policy"),
        ("unsigned long *", "nmask"),
        ("unsigned long", "maxnode"),
        ("unsigned long", "addr"),
        ("unsigned long", "flags"),
    )),
    "get_robust_list": ("long", (
        ("int", "pid"),
        ("struct robust_list_head * *", "head_ptr"),
        ("size_t *", "len_ptr"),
    )),
    "getcpu": ("long", (
        ("unsigned *", "cpu"),
        ("unsigned *", "node"),
        ("struct getcpu_cache *", "cache"),
    )),
    "getcwd": ("long", (
        ("char *", "pathname"),
        ("unsigned long", "size"),
    )),
    "getdents": ("long", (
        ("unsigned int", "fd"),
        ("struct linux_dirent *", "dirent"),
        ("unsigned int", "count"),
    )),
    "getdents64": ("long", (
        ("unsigned int", "fd"),
        ("struct linux_dirent64 *", "dirent"),
        ("unsigned int", "count"),
    )),
    "getegid": ("long", (
    )),
    "getegid16": ("long", (
    )),
    "geteuid": ("long", (
    )),
    "geteuid16": ("long", (
    )),
    "getgid": ("long", (
    )),
    "getgid16": ("long", (
    )),
    "getgroups": ("long", (
        ("int", "gidsetsize"),
        ("gid_t *", "grouplist"),
    )),
    "getgroups16": ("long", (
        ("int", "gidsetsize"),
        ("old_gid_t *", "grouplist"),
    )),
    "gethostname": ("long", (
        ("char *", "name"),
        ("int", "len"),
    )),
    "getitimer": ("long", (
        ("int", "which"),
        ("struct itimerval *", "value"),
    )),
    "getpeername": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int *", "addrlen"),
    )),
    "getpgid": ("long", (
        ("pid_t", "pid"),
    )),
    "getpgrp": ("long", (
    )),
    "getpid": ("long", (
    )),
    "getppid": ("long", (
    )),
    "getpriority": ("long", (
        ("int", "which"),
        ("int", "who"),
    )),
    "getrandom": ("long", (
        ("char *", "buf"),
        ("size_t", "count"),
        ("unsigned int", "flags"),
    )),
    "getresgid": ("long", (
        ("gid_t *", "rgid"),
        ("gid_t *", "egid"),
        ("gid_t *", "sgid"),
    )),
    "getresgid16": ("long", (
        ("old_gid_t *", "rgid"),
        ("old_gid_t *", "egid"),
        ("old_gid_t *", "sgid"),
    )),
    "getresuid": ("long", (
        ("uid_t *", "ruid"),
        ("uid_t *", "euid"),
        ("uid_t *", "suid"),
    )),
    "getresuid16": ("long", (
        ("old_uid_t *", "ruid"),
        ("old_uid_t *", "euid"),
        ("old_uid_t *", "suid"),
    )),
    "getrlimit": ("long", (
        ("unsigned int", "resource"),
        ("struct rlimit *", "rlim"),
    )),
    "getrusage": ("long", (
        ("int", "who"),
        ("struct rusage *", "ru"),
    )),
    "getsid": ("long", (
        ("pid_t", "pid"),
    )),
    "getsockname": ("long", (
        ("int", "sockfd"),
        ("struct sockaddr *", "addr"),
        ("int *", "addrlen"),
    )),
    "getsockopt": ("long", (
        ("int", "fd"),
        ("int", "level"),
        ("int", "optname"),
        ("char *", "optval"),
        ("int *", "optlen"),
    )),
    "gettid": ("long", (
    )),
    "gettimeofday": ("long", (
        ("struct timeval *", "tv"),
        ("struct timezone *", "tz"),
    )),
    "getuid": ("long", (
    )),
    "getuid16": ("long", (
    )),
    "getxattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
        ("void *", "value"),
        ("size_t", "size"),
    )),
    "init_module": ("long", (
        ("void *", "umod"),
        ("unsigned long", "len"),
        ("const char *", "uargs"),
    )),
    "inotify_add_watch": ("long", (
        ("int", "fd"),
        ("const char *", "path"),
        ("u32", "mask"),
    )),
    "inotify_init": ("long", (
    )),
    "inotify_init1": ("long", (
        ("int", "flags"),
    )),
    "inotify_rm_watch": ("long", (
        ("int", "fd"),
        ("__s32", "wd"),
    )),
    "io_cancel": ("long", (
        ("aio_context_t", "ctx_id"),
        ("struct iocb *", "iocb"),
        ("struct io_event *", "result"),
    )),
    "io_destroy": ("long", (
        ("aio_context_t", "ctx"),
    )),
    "io_getevents": ("long", (
        ("aio_context_t", "ctx_id"),
        ("long", "min_nr"),
        ("long", "nr"),
        ("struct io_event *", "events"),
        ("struct timespec *", "timeout"),
    )),
    "io_setup": ("long", (
        ("unsigned", "nr_reqs"),
        ("aio_context_t *", "ctx"),
    )),
    "io_submit": ("long", (
        ("aio_context_t", "ctx_id"),
        ("long", "nr"),
        ("struct iocb * *", "iocbpp"),
    )),
    "ioctl": ("long", (
        ("unsigned int", "fd"),
        ("unsigned int", "cmd"),
        ("unsigned long", "arg"),
    )),
    "ioperm": ("long", (
        ("unsigned long", "from"),
        ("unsigned long", "num"),
        ("int", "on"),
    )),
    "ioprio_get": ("long", (
        ("int", "which"),
        ("int", "who"),
    )),
    "ioprio_set": ("long", (
        ("int", "which"),
        ("int", "who"),
        ("int", "ioprio"),
    )),
    "ipc": ("long", (
        ("unsigned int", "call"),
        ("int", "first"),
        ("unsigned long", "second"),
        ("unsigned long", "third"),
        ("void *", "ptr"),
        ("long", "fifth"),
    )),
    "kcmp": ("long", (
        ("pid_t", "pid1"),
        ("pid_t", "pid2"),
        ("int", "type"),
        ("unsigned long", "idx1"),
        ("unsigned long", "idx2"),
    )),
    "kexec_file_load": ("long", (
        ("int", "kernel_fd"),
        ("int", "initrd_fd"),
        ("unsigned long", "cmdline_len"),
        ("const char *", "cmdline_ptr"),
        ("unsigned long", "flags"),
    )),
    "kexec_load": ("long", (
        ("unsigned long", "entry"),
        ("unsigned long", "nr_segments"),
        ("struct kexec_segment *", "segments"),
        ("unsigned long", "flags"),
    )),
    "keyctl": ("long", (
        ("int", "cmd"),
        ("unsigned long", "arg2"),
        ("unsigned long", "arg3"),
        ("unsigned long", "arg4"),
        ("unsigned long", "arg5"),
    )),
    "kill": ("long", (
        ("int", "pid"),
        ("int", "sig"),
    )),
    "lchown": ("long", (
        ("const char *", "filename"),
        ("uid_t", "user"),
        ("gid_t", "group"),
    )),
    "lchown16": ("long", (
        ("const char *", "filename"),
        ("old_uid_t", "user"),
        ("old_gid_t", "group"),
    )),
    "lgetxattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
        ("void *", "value"),
        ("size_t", "size"),
    )),
    "link": ("long", (
        ("const char *", "oldname"),
        ("const char *", "newname"),
    )),
    "linkat": ("long", (
        ("int", "olddfd"),
        ("const char *", "oldname"),
        ("int", "newdfd"),
        ("const char *", "newname"),
        ("int", "flags"),
    )),
    "listen": ("long", (
        ("int", "sockfd"),
        ("int", "backlog"),
    )),
    "listxattr": ("long", (
        ("const char *", "path"),
        ("char *", "list"),
        ("size_t", "size"),
    )),
    "llistxattr": ("long", (
        ("const char *", "path"),
        ("char *", "list"),
        ("size_t", "size"),
    )),
    "llseek": ("long", (
        ("unsigned int", "fd"),
        ("unsigned long", "offset_high"),
        ("unsigned long", "offset_low"),
        ("loff_t *", "result"),
        ("unsigned int", "whence"),
    )),
    "lookup_dcookie": ("long", (
        ("u64", "cookie64"),
        ("char *", "buf"),
        ("size_t", "len"),
    )),
    "lremovexattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
    )),
    "lseek": ("long", (
        ("unsigned int", "fd"),
        ("off_t", "offset"),
        ("unsigned int", "whence"),
    )),
    "lsetxattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
        ("const void *", "value"),
        ("size_t", "size"),
        ("int", "flags"),
    )),
    "lstat": ("long", (
        ("const char *", "filename"),
        ("struct __old_kernel_stat *", "statbuf"),
    )),
    "lstat64": ("long", (
        ("const char *", "filename"),
        ("struct stat64 *", "statbuf"),
    )),
    "madvise": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
        ("int", "behavior"),
    )),
    "mbind": ("long", (
        ("unsigned long", "start"),
        ("unsigned long", "len"),
        ("unsigned long", "mode"),
        ("const unsigned long *", "nmask"),
        ("unsigned long", "maxnode"),
        ("unsigned", "flags"),
    )),
    "membarrier": ("long", (
        ("int", "cmd"),
        ("int", "flags"),
    )),
    "memfd_create": ("long", (
        ("const char *", "uname_ptr"),
        ("unsigned int", "flags"),
    )),
    "migrate_pages": ("long", (
        ("pid_t", "pid"),
        ("unsigned long", "maxnode"),
        ("const unsigned long *", "from"),
        ("const unsigned long *", "to"),
    )),
    "mincore": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
        ("unsigned char *", "vec"),
    )),
    "mkdir": ("long", (
        ("const char *", "pathname"),
        ("umode_t", "mode"),
    )),
    "mkdirat": ("long", (
        ("int", "dfd"),
        ("const char *", "pathname"),
        ("umode_t", "mode"),
    )),
    "mknod": ("long", (
        ("const char *", "filename"),
        ("umode_t", "mode"),
        ("unsigned", "dev"),
    )),
    "mknodat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("umode_t", "mode"),
        ("unsigned", "dev"),
    )),
    "mlock": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
    )),
    "mlock2": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
        ("int", "flags"),
    )),
    "mlockall": ("long", (
        ("int", "flags"),
    )),
    "mmap_pgoff": ("long", (
        ("unsigned long", "addr"),
        ("unsigned long", "len"),
        ("unsigned long", "prot"),
        ("unsigned long", "flags"),
        ("unsigned long", "fd"),
        ("unsigned long", "pgoff"),
    )),
    "mount": ("long", (
        ("char *", "dev_name"),
        ("char *", "dir_name"),
        ("char *", "type"),
        ("unsigned long", "flags"),
        ("void *", "data"),
    )),
    "move_pages": ("long", (
        ("pid_t", "pid"),
        ("unsigned long", "nr_pages"),
        ("const void * *", "pages"),
        ("const int *", "nodes"),
        ("int *", "status"),
        ("int", "flags"),
    )),
    "mprotect": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
        ("unsigned long", "prot"),
    )),
    "mq_getsetattr": ("long", (
        ("mqd_t", "mqdes"),
        ("const struct mq_attr *", "mqstat"),
        ("struct mq_attr *", "omqstat"),
    )),
    "mq_notify": ("long", (
        ("mqd_t", "mqdes"),
        ("const struct sigevent *", "notification"),
    )),
    "mq_open": ("long", (
        ("const char *", "name"),
        ("int", "oflag"),
        ("umode_t", "mode"),
        ("struct mq_attr *", "attr"),
    )),
    "mq_timedreceive": ("long", (
        ("mqd_t", "mqdes"),
        ("char *", "msg_ptr"),
        ("size_t", "msg_len"),
        ("unsigned int *", "msg_prio"),
        ("const struct timespec *", "abs_timeout"),
    )),
    "mq_timedsend": ("long", (
        ("mqd_t", "mqdes"),
        ("const char *", "msg_ptr"),
        ("size_t", "msg_len"),
        ("unsigned int", "msg_prio"),
        ("const struct timespec *", "abs_timeout"),
    )),
    "mq_unlink": ("long", (
        ("const char *", "name"),
    )),
    "mremap": ("long", (
        ("unsigned long", "addr"),
        ("unsigned long", "old_len"),
        ("unsigned long", "new_len"),
        ("unsigned long", "flags"),
        ("unsigned long", "new_addr"),
    )),
    "msgctl": ("long", (
        ("int", "msqid"),
        ("int", "cmd"),
        ("struct msqid_ds *", "buf"),
    )),
    "msgget": ("long", (
        ("key_t", "key"),
        ("int", "msgflg"),
    )),
    "msgrcv": ("long", (
        ("int", "msqid"),
        ("struct msgbuf *", "msgp"),
        ("size_t", "msgsz"),
        ("long", "msgtyp"),
        ("int", "msgflg"),
    )),
    "msgsnd": ("long", (
        ("int", "msqid"),
        ("struct msgbuf *", "msgp"),
        ("size_t", "msgsz"),
        ("int", "msgflg"),
    )),
    "msync": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
        ("int", "flags"),
    )),
    "munlock": ("long", (
        ("unsigned long", "start"),
        ("size_t", "len"),
    )),
    "munlockall": ("long", (
    )),
    "munmap": ("long", (
        ("unsigned long", "addr"),
        ("size_t", "len"),
    )),
    "name_to_handle_at": ("long", (
        ("int", "dirfd"),
        ("const char *", "name"),
        ("struct file_handle *", "handle"),
        ("int *", "mnt_id"),
        ("int", "flag"),
    )),
    "nanosleep": ("long", (
        ("struct timespec *", "rqtp"),
        ("struct timespec *", "rmtp"),
    )),
    "newfstat": ("long", (
        ("unsigned int", "fd"),
        ("struct stat *", "statbuf"),
    )),
    "newfstatat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("struct stat *", "statbuf"),
        ("int", "flag"),
    )),
    "newlstat": ("long", (
        ("const char *", "filename"),
        ("struct stat *", "statbuf"),
    )),
    "newstat": ("long", (
        ("const char *", "filename"),
        ("struct stat *", "statbuf"),
    )),
    "newuname": ("long", (
        ("struct new_utsname *", "name"),
    )),
    "ni_syscall": ("long", (
    )),
    "nice": ("long", (
        ("int", "increment"),
    )),
    "old_getrlimit": ("long", (
        ("unsigned int", "resource"),
        ("struct rlimit *", "rlim"),
    )),
    "old_mmap": ("long", (
        ("struct mmap_arg_struct *", "arg"),
    )),
    "old_readdir": ("long", (
        ("unsigned int", "fd"),
        ("struct old_linux_dirent *", "dirp"),
        ("unsigned int", "count"),
    )),
    "old_select": ("long", (
        ("struct sel_arg_struct *", "arg"),
    )),
    "oldumount": ("long", (
        ("char *", "name"),
    )),
    "olduname": ("long", (
        ("struct oldold_utsname *", "buf"),
    )),
    "open": ("long", (
        ("const char *", "filename"),
        ("int", "flags"),
        ("umode_t", "mode"),
    )),
    "open_by_handle_at": ("long", (
        ("int", "mount_fd"),
        ("struct file_handle *", "handle"),
        ("int", "flags"),
    )),
    "openat": ("long", (
        ("int", "dirfd"),
        ("const char *", "filename"),
        ("int", "flags"),
        ("umode_t", "mode"),
    )),
    "pause": ("long", (
    )),
    "pciconfig_iobase": ("long", (
        ("long", "which"),
        ("unsigned long", "bus"),
        ("unsigned long", "devfn"),
    )),
    "pciconfig_read": ("long", (
        ("unsigned long", "bus"),
        ("unsigned long", "dfn"),
        ("unsigned long", "off"),
        ("unsigned long", "len"),
        ("void *", "buf"),
    )),
    "pciconfig_write": ("long", (
        ("unsigned long", "bus"),
        ("unsigned long", "dfn"),
        ("unsigned long", "off"),
        ("unsigned long", "len"),
        ("void *", "buf"),
    )),
    "perf_event_open": ("long", (
        ("struct perf_event_attr *", "attr_uptr"),
        ("pid_t", "pid"),
        ("int", "cpu"),
        ("int", "group_fd"),
        ("unsigned long", "flags"),
    )),
    "personality": ("long", (
        ("unsigned int", "personality"),
    )),
    "pipe": ("long", (
        ("int *", "fildes"),
    )),
    "pipe2": ("long", (
        ("int *", "fildes"),
        ("int", "flags"),
    )),
    "pivot_root": ("long", (
        ("const char *", "new_root"),
        ("const char *", "put_old"),
    )),
    "poll": ("long", (
        ("struct pollfd *", "ufds"),
        ("unsigned int", "nfds"),
        ("int", "timeout"),
    )),
    "ppoll": ("long", (
        ("struct pollfd *", "fds"),
        ("unsigned int", "nfds"),
        ("struct timespec *", "tmo_p"),
        ("const sigset_t *", "sigmask"),
        ("size_t", ""),
    )),
    "prctl": ("long", (
        ("int", "option"),
        ("unsigned long", "arg2"),
        ("unsigned long", "arg3"),
        ("unsigned long", "arg4"),
        ("unsigned long", "arg5"),
    )),
    "pread64": ("long", (
        ("unsigned int", "fd"),
        ("char *", "buf"),
        ("size_t", "count"),
        ("loff_t", "pos"),
    )),
    "preadv": ("long", (
        ("unsigned long", "fd"),
        ("const struct iovec *", "vec"),
        ("unsigned long", "vlen"),
        ("unsigned long", "pos_l"),
        ("unsigned long", "pos_h"),
    )),
    "prlimit64": ("long", (
        ("pid_t", "pid"),
        ("unsigned int", "resource"),
        ("const struct rlimit64 *", "new_rlim"),
        ("struct rlimit64 *", "old_rlim"),
    )),
    "process_vm_readv": ("long", (
        ("pid_t", "pid"),
        ("const struct iovec *", "lvec"),
        ("unsigned long", "liovcnt"),
        ("const struct iovec *", "rvec"),
        ("unsigned long", "riovcnt"),
        ("unsigned long", "flags"),
    )),
    "process_vm_writev": ("long", (
        ("pid_t", "pid"),
        ("const struct iovec *", "lvec"),
        ("unsigned long", "liovcnt"),
        ("const struct iovec *", "rvec"),
        ("unsigned long", "riovcnt"),
        ("unsigned long", "flags"),
    )),
    "pselect6": ("long", (
        ("int", "nfds"),
        ("fd_set *", "readfds"),
        ("fd_set *", "writefds"),
        ("fd_set *", "exceptfds"),
        ("struct timespec *", "timeout"),
        ("void *", "sigmask"),
    )),
    "ptrace": ("long", (
        ("long", "request"),
        ("long", "pid"),
        ("unsigned long", "addr"),
        ("unsigned long", "data"),
    )),
    "pwrite64": ("long", (
        ("unsigned int", "fd"),
        ("const char *", "buf"),
        ("size_t", "count"),
        ("loff_t", "pos"),
    )),
    "pwritev": ("long", (
        ("unsigned long", "fd"),
        ("const struct iovec *", "vec"),
        ("unsigned long", "vlen"),
        ("unsigned long", "pos_l"),
        ("unsigned long", "pos_h"),
    )),
    "quotactl": ("long", (
        ("unsigned int", "cmd"),
        ("const char *", "special"),
        ("qid_t", "id"),
        ("void *", "addr"),
    )),
    "read": ("long", (
        ("unsigned int", "fd"),
        ("char *", "buf"),
        ("size_t", "count"),
    )),
    "readahead": ("long", (
        ("int", "fd"),
        ("loff_t", "offset"),
        ("size_t", "count"),
    )),
    "readlink": ("long", (
        ("const char *", "path"),
        ("char *", "buf"),
        ("int", "bufsiz"),
    )),
    "readlinkat": ("long", (
        ("int", "dfd"),
        ("const char *", "path"),
        ("char *", "buf"),
        ("int", "bufsiz"),
    )),
    "readv": ("long", (
        ("unsigned long", "fd"),
        ("const struct iovec *", "vec"),
        ("unsigned long", "vlen"),
    )),
    "reboot": ("long", (
        ("int", "magic1"),
        ("int", "magic2"),
        ("unsigned int", "cmd"),
        ("void *", "arg"),
    )),
    "recv": ("long", (
        ("int", "sockfd"),
        ("void *", "buf"),
        ("size_t", "len"),
        ("unsigned", "flags"),
    )),
    "recvfrom": ("long", (
        ("int", "sockfd"),
        ("void *", "buf"),
        ("size_t", "len"),
        ("unsigned", "flags"),
        ("struct sockaddr *", "src_addr"),
        ("int *", "addrlen"),
    )),
    "recvmmsg": ("long", (
        ("int", "fd"),
        ("struct mmsghdr *", "msg"),
        ("unsigned int", "vlen"),
        ("unsigned", "flags"),
        ("struct timespec *", "timeout"),
    )),
    "recvmsg": ("long", (
        ("int", "fd"),
        ("struct user_msghdr *", "msg"),
        ("unsigned", "flags"),
    )),
    "remap_file_pages": ("long", (
        ("unsigned long", "start"),
        ("unsigned long", "size"),
        ("unsigned long", "prot"),
        ("unsigned long", "pgoff"),
        ("unsigned long", "flags"),
    )),
    "removexattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
    )),
    "rename": ("long", (
        ("const char *", "oldname"),
        ("const char *", "newname"),
    )),
    "renameat": ("long", (
        ("int", "olddfd"),
        ("const char *", "oldname"),
        ("int", "newdfd"),
        ("const char *", "newname"),
    )),
    "renameat2": ("long", (
        ("int", "olddfd"),
        ("const char *", "oldname"),
        ("int", "newdfd"),
        ("const char *", "newname"),
        ("unsigned int", "flags"),
    )),
    "request_key": ("long", (
        ("const char *", "_type"),
        ("const char *", "_description"),
        ("const char *", "_callout_info"),
        ("key_serial_t", "destringid"),
    )),
    "restart_syscall": ("long", (
    )),
    "rmdir": ("long", (
        ("const char *", "pathname"),
    )),
    "rt_sigaction": ("long", (
        ("int", "signum"),
        ("const struct sigaction *", "act"),
        ("struct sigaction *", "oldact"),
        ("size_t", ""),
    )),
    "rt_sigpending": ("long", (
        ("sigset_t *", "set"),
        ("size_t", "sigsetsize"),
    )),
    "rt_sigprocmask": ("long", (
        ("int", "how"),
        ("sigset_t *", "set"),
        ("sigset_t *", "oset"),
        ("size_t", "sigsetsize"),
    )),
    "rt_sigqueueinfo": ("long", (
        ("int", "pid"),
        ("int", "sig"),
        ("siginfo_t *", "uinfo"),
    )),
    "rt_sigsuspend": ("long", (
        ("sigset_t *", "unewset"),
        ("size_t", "sigsetsize"),
    )),
    "rt_sigtimedwait": ("long", (
        ("const sigset_t *", "uthese"),
        ("siginfo_t *", "uinfo"),
        ("const struct timespec *", "uts"),
        ("size_t", "sigsetsize"),
    )),
    "rt_tgsigqueueinfo": ("long", (
        ("pid_t", "tgid"),
        ("pid_t", "pid"),
        ("int", "sig"),
        ("siginfo_t *", "uinfo"),
    )),
    "sched_get_priority_max": ("long", (
        ("int", "policy"),
    )),
    "sched_get_priority_min": ("long", (
        ("int", "policy"),
    )),
    "sched_getaffinity": ("long", (
        ("pid_t", "pid"),
        ("unsigned int", "len"),
        ("unsigned long *", "user_mask_ptr"),
    )),
    "sched_getattr": ("long", (
        ("pid_t", "pid"),
        ("struct sched_attr *", "attr"),
        ("unsigned int", "size"),
        ("unsigned int", "flags"),
    )),
    "sched_getparam": ("long", (
        ("pid_t", "pid"),
        ("struct sched_param *", "param"),
    )),
    "sched_getscheduler": ("long", (
        ("pid_t", "pid"),
    )),
    "sched_rr_get_interval": ("long", (
        ("pid_t", "pid"),
        ("struct timespec *", "interval"),
    )),
    "sched_setaffinity": ("long", (
        ("pid_t", "pid"),
        ("unsigned int", "len"),
        ("unsigned long *", "user_mask_ptr"),
    )),
    "sched_setattr": ("long", (
        ("pid_t", "pid"),
        ("struct sched_attr *", "attr"),
        ("unsigned int", "flags"),
    )),
    "sched_setparam": ("long", (
        ("pid_t", "pid"),
        ("struct sched_param *", "param"),
    )),
    "sched_setscheduler": ("long", (
        ("pid_t", "pid"),
        ("int", "policy"),
        ("struct sched_param *", "param"),
    )),
    "sched_yield": ("long", (
    )),
    "seccomp": ("long", (
        ("unsigned int", "op"),
        ("unsigned int", "flags"),
        ("const char *", "uargs"),
    )),
    "select": ("long", (
        ("int", "nfds"),
        ("fd_set *", "readfds"),
        ("fd_set *", "writefds"),
        ("fd_set *", "errorfds"),
        ("struct timeval *", "timeout"),
    )),
    "semctl": ("long", (
        ("int", "semid"),
        ("int", "semnum"),
        ("int", "cmd"),
        ("unsigned long", "arg"),
    )),
    "semget": ("long", (
        ("key_t", "key"),
        ("int", "nsems"),
        ("int", "semflg"),
    )),
    "semop": ("long", (
        ("int", "semid"),
        ("struct sembuf *", "sops"),
        ("unsigned", "nsops"),
    )),
    "semtimedop": ("long", (
        ("int", "semid"),
        ("struct sembuf *", "sops"),
        ("unsigned", "nsops"),
        ("const struct timespec *", "timeout"),
    )),
    "send": ("long", (
        ("int", "sockfd"),
        ("void *", "buf"),
        ("size_t", "len"),
        ("unsigned", "flags"),
    )),
    "sendfile": ("long", (
        ("int", "out_fd"),
        ("int", "in_fd"),
        ("off_t *", "offset"),
        ("size_t", "count"),
    )),
    "sendfile64": ("long", (
        ("int", "out_fd"),
        ("int", "in_fd"),
        ("loff_t *", "offset"),
        ("size_t", "count"),
    )),
    "sendmmsg": ("long", (
        ("int", "fd"),
        ("struct mmsghdr *", "msg"),
        ("unsigned int", "vlen"),
        ("unsigned", "flags"),
    )),
    "sendmsg": ("long", (
        ("int", "fd"),
        ("struct user_msghdr *", "msg"),
        ("unsigned", "flags"),
    )),
    "sendto": ("long", (
        ("int", "sockfd"),
        ("void *", "buf"),
        ("size_t", "len"),
        ("unsigned", "flags"),
        ("struct sockaddr *", "dest_addr"),
        ("int", "addrlen"),
    )),
    "set_mempolicy": ("long", (
        ("int", "mode"),
        ("const unsigned long *", "nmask"),
        ("unsigned long", "maxnode"),
    )),
    "set_robust_list": ("long", (
        ("struct robust_list_head *", "head"),
        ("size_t", "len"),
    )),
    "set_tid_address": ("long", (
        ("int *", "tidptr"),
    )),
    "setdomainname": ("long", (
        ("char *", "name"),
        ("int", "len"),
    )),
    "setfsgid": ("long", (
        ("gid_t", "gid"),
    )),
    "setfsgid16": ("long", (
        ("old_gid_t", "gid"),
    )),
    "setfsuid": ("long", (
        ("uid_t", "uid"),
    )),
    "setfsuid16": ("long", (
        ("old_uid_t", "uid"),
    )),
    "setgid": ("long", (
        ("gid_t", "gid"),
    )),
    "setgid16": ("long", (
        ("old_gid_t", "gid"),
    )),
    "setgroups": ("long", (
        ("int", "gidsetsize"),
        ("gid_t *", "grouplist"),
    )),
    "setgroups16": ("long", (
        ("int", "gidsetsize"),
        ("old_gid_t *", "grouplist"),
    )),
    "sethostname": ("long", (
        ("char *", "name"),
        ("int", "len"),
    )),
    "setitimer": ("long", (
        ("int", "which"),
        ("struct itimerval *", "value"),
        ("struct itimerval *", "ovalue"),
    )),
    "setns": ("long", (
        ("int", "fd"),
        ("int", "nstype"),
    )),
    "setpgid": ("long", (
        ("pid_t", "pid"),
        ("pid_t", "pgid"),
    )),
    "setpriority": ("long", (
        ("int", "which"),
        ("int", "who"),
        ("int", "niceval"),
    )),
    "setregid": ("long", (
        ("gid_t", "rgid"),
        ("gid_t", "egid"),
    )),
    "setregid16": ("long", (
        ("old_gid_t", "rgid"),
        ("old_gid_t", "egid"),
    )),
    "setresgid": ("long", (
        ("gid_t", "rgid"),
        ("gid_t", "egid"),
        ("gid_t", "sgid"),
    )),
    "setresgid16": ("long", (
        ("old_gid_t", "rgid"),
        ("old_gid_t", "egid"),
        ("old_gid_t", "sgid"),
    )),
    "setresuid": ("long", (
        ("uid_t", "ruid"),
        ("uid_t", "euid"),
        ("uid_t", "suid"),
    )),
    "setresuid16": ("long", (
        ("old_uid_t", "ruid"),
        ("old_uid_t", "euid"),
        ("old_uid_t", "suid"),
    )),
    "setreuid": ("long", (
        ("uid_t", "ruid"),
        ("uid_t", "euid"),
    )),
    "setreuid16": ("long", (
        ("old_uid_t", "ruid"),
        ("old_uid_t", "euid"),
    )),
    "setrlimit": ("long", (
        ("unsigned int", "resource"),
        ("struct rlimit *", "rlim"),
    )),
    "setsid": ("long", (
    )),
    "setsockopt": ("long", (
        ("int", "fd"),
        ("int", "level"),
        ("int", "optname"),
        ("char *", "optval"),
        ("int", "optlen"),
    )),
    "settimeofday": ("long", (
        ("struct timeval *", "tv"),
        ("struct timezone *", "tz"),
    )),
    "setuid": ("long", (
        ("uid_t", "uid"),
    )),
    "setuid16": ("long", (
        ("old_uid_t", "uid"),
    )),
    "setxattr": ("long", (
        ("const char *", "path"),
        ("const char *", "name"),
        ("const void *", "value"),
        ("size_t", "size"),
        ("int", "flags"),
    )),
    "sgetmask": ("long", (
    )),
    "shmat": ("long", (
        ("int", "shmid"),
        ("char *", "shmaddr"),
        ("int", "shmflg"),
    )),
    "shmctl": ("long", (
        ("int", "shmid"),
        ("int", "cmd"),
        ("struct shmid_ds *", "buf"),
    )),
    "shmdt": ("long", (
        ("char *", "shmaddr"),
    )),
    "shmget": ("long", (
        ("key_t", "key"),
        ("size_t", "size"),
        ("int", "flag"),
    )),
    "shutdown": ("long", (
        ("int", "sockfd"),
        ("int", "how"),
    )),
    "sigaction": ("long", (
        ("int", "signum"),
        ("const struct old_sigaction *", "act"),
        ("struct old_sigaction *", "oldact"),
    )),
    "sigaltstack": ("long", (
        ("const struct sigaltstack *", "uss"),
        ("struct sigaltstack *", "uoss"),
    )),
    "signal": ("long", (
        ("int", "sig"),
        ("__sighandler_t", "handler"),
    )),
    "signalfd": ("long", (
        ("int", "ufd"),
        ("sigset_t *", "user_mask"),
        ("size_t", "sizemask"),
    )),
    "signalfd4": ("long", (
        ("int", "ufd"),
        ("sigset_t *", "user_mask"),
        ("size_t", "sizemask"),
        ("int", "flags"),
    )),
    "sigpending": ("long", (
        ("old_sigset_t *", "set"),
    )),
    "sigprocmask": ("long", (
        ("int", "how"),
        ("old_sigset_t *", "set"),
        ("old_sigset_t *", "oset"),
    )),
    "sigsuspend": ("long", (
        ("int", "unused1"),
        ("int", "unused2"),
        ("old_sigset_t", "mask"),
    )),
    "socket": ("long", (
        ("int", "domain"),
        ("int", "type"),
        ("int", "protocol"),
    )),
    "socketcall": ("long", (
        ("int", "call"),
        ("unsigned long *", "args"),
    )),
    "socketpair": ("long", (
        ("int", "domain"),
        ("int", "type"),
        ("int", "protocol"),
        ("int *", "sv"),
    )),
    "splice": ("long", (
        ("int", "fd_in"),
        ("loff_t *", "off_in"),
        ("int", "fd_out"),
        ("loff_t *", "off_out"),
        ("size_t", "len"),
        ("unsigned int", "flags"),
    )),
    "spu_create": ("long", (
        ("const char *", "name"),
        ("unsigned int", "flags"),
        ("umode_t", "mode"),
        ("int", "fd"),
    )),
    "spu_run": ("long", (
        ("int", "fd"),
        ("__u32 *", "unpc"),
        ("__u32 *", "ustatus"),
    )),
    "ssetmask": ("long", (
        ("int", "newmask"),
    )),
    "stat": ("long", (
        ("const char *", "filename"),
        ("struct __old_kernel_stat *", "statbuf"),
    )),
    "stat64": ("long", (
        ("const char *", "filename"),
        ("struct stat64 *", "statbuf"),
    )),
    "statfs": ("long", (
        ("const char *", "path"),
        ("struct statfs *", "buf"),
    )),
    "statfs64": ("long", (
        ("const char *", "path"),
        ("size_t", "sz"),
        ("struct statfs64 *", "buf"),
    )),
    "statx": ("long", (
        ("int", "dirfd"),
        ("const char *", "pathname"),
        ("int", "flags"),
        ("unsigned int", "mask"),
        ("struct statx *", "statxbuf"),
    )),
    "stime": ("long", (
        ("time_t *", "tptr"),
    )),
    "swapoff": ("long", (
        ("const char *", "specialfile"),
    )),
    "swapon": ("long", (
        ("const char *", "specialfile"),
        ("int", "swap_flags"),
    )),
    "symlink": ("long", (
        ("const char *", "old"),
        ("const char *", "new"),
    )),
    "symlinkat": ("long", (
        ("const char *", "oldname"),
        ("int", "newdfd"),
        ("const char *", "newname"),
    )),
    "sync": ("long", (
    )),
    "sync_file_range": ("long", (
        ("int", "fd"),
        ("loff_t", "offset"),
        ("loff_t", "nbytes"),
        ("unsigned int", "flags"),
    )),
    "sync_file_range2": ("long", (
        ("int", "fd"),
        ("unsigned int", "flags"),
        ("loff_t", "offset"),
        ("loff_t", "nbytes"),
    )),
    "syncfs": ("long", (
        ("int", "fd"),
    )),
    "sysctl": ("long", (
        ("struct __sysctl_args *", "args"),
    )),
    "sysfs": ("long", (
        ("int", "option"),
        ("unsigned long", "arg1"),
        ("unsigned long", "arg2"),
    )),
    "sysinfo": ("long", (
        ("struct sysinfo *", "info"),
    )),
    "syslog": ("long", (
        ("int", "type"),
        ("char *", "buf"),
        ("int", "len"),
    )),
    "tee": ("long", (
        ("int", "fdin"),
        ("int", "fdout"),
        ("size_t", "len"),
        ("unsigned int", "flags"),
    )),
    "tgkill": ("long", (
        ("int", "tgid"),
        ("int", "pid"),
        ("int", "sig"),
    )),
    "time": ("long", (
        ("time_t *", "tloc"),
    )),
    "timer_create": ("long", (
        ("clockid_t", "which_clock"),
        ("struct sigevent *", "timer_event_spec"),
        ("timer_t *", "created_timer_id"),
    )),
    "timer_delete": ("long", (
        ("timer_t", "timer_id"),
    )),
    "timer_getoverrun": ("long", (
        ("timer_t", "timer_id"),
    )),
    "timer_gettime": ("long", (
        ("timer_t", "timer_id"),
        ("struct itimerspec *", "setting"),
    )),
    "timer_settime": ("long", (
        ("timer_t", "timer_id"),
        ("int", "flags"),
        ("const struct itimerspec *", "new_setting"),
        ("struct itimerspec *", "old_setting"),
    )),
    "timerfd_create": ("long", (
        ("int", "clockid"),
        ("int", "flags"),
    )),
    "timerfd_gettime": ("long", (
        ("int", "ufd"),
        ("struct itimerspec *", "otmr"),
    )),
    "timerfd_settime": ("long", (
        ("int", "ufd"),
        ("int", "flags"),
        ("const struct itimerspec *", "utmr"),
        ("struct itimerspec *", "otmr"),
    )),
    "times": ("long", (
        ("struct tms *", "tbuf"),
    )),
    "tkill": ("long", (
        ("int", "pid"),
        ("int", "sig"),
    )),
    "truncate": ("long", (
        ("const char *", "path"),
        ("long", "length"),
    )),
    "truncate64": ("long", (
        ("const char *", "path"),
        ("loff_t", "length"),
    )),
    "umask": ("long", (
        ("int", "mask"),
    )),
    "umount": ("long", (
        ("char *", "name"),
        ("int", "flags"),
    )),
    "uname": ("long", (
        ("struct old_utsname *", "buf"),
    )),
    "unlink": ("long", (
        ("const char *", "pathname"),
    )),
    "unlinkat": ("long", (
        ("int", "dfd"),
        ("const char *", "pathname"),
        ("int", "flag"),
    )),
    "unshare": ("long", (
        ("unsigned long", "unshare_flags"),
    )),
    "uselib": ("long", (
        ("const char *", "library"),
    )),
    "userfaultfd": ("long", (
        ("int", "flags"),
    )),
    "ustat": ("long", (
        ("unsigned", "dev"),
        ("struct ustat *", "ubuf"),
    )),
    "utime": ("long", (
        ("char *", "filename"),
        ("struct utimbuf *", "times"),
    )),
    "utimensat": ("long", (
        ("int", "dfd"),
        ("const char *", "filename"),
        ("struct timespec *", "utimes"),
        ("int", "flags"),
    )),
    "utimes": ("long", (
        ("char *", "filename"),
        ("struct timeval *", "utimes"),
    )),
    "vfork": ("long", (
    )),
    "vhangup": ("long", (
    )),
    "vmsplice": ("long", (
        ("int", "fd"),
        ("const struct iovec *", "iov"),
        ("unsigned long", "nr_segs"),
        ("unsigned int", "flags"),
    )),
    "wait4": ("long", (
        ("pid_t", "pid"),
        ("int *", "stat_addr"),
        ("int", "options"),
        ("struct rusage *", "ru"),
    )),
    "waitid": ("long", (
        ("int", "which"),
        ("pid_t", "pid"),
        ("struct siginfo *", "infop"),
        ("int", "options"),
        ("struct rusage *", "ru"),
    )),
    "waitpid": ("long", (
        ("pid_t", "pid"),
        ("int *", "stat_addr"),
        ("int", "options"),
    )),
    "write": ("long", (
        ("unsigned int", "fd"),
        ("const char *", "buf"),
        ("size_t", "count"),
    )),
    "writev": ("long", (
        ("unsigned long", "fd"),
        ("const struct iovec *", "vec"),
        ("unsigned long", "vlen"),
    )),
}

for orig, copies in ALIASES.items():
    orig = SYSCALL_PROTOTYPES[orig]
    for copy in copies:
        SYSCALL_PROTOTYPES[copy] = orig



from ctypes import sizeof
from ptrace.ctypes_tools import formatUintHex16, formatUintHex32, formatWordHex
from datetime import datetime, timedelta
from os import getenv, access, X_OK, pathsep, getcwd
from os.path import join as path_join, isabs, dirname, normpath


def dumpRegs(log, regs):
    """
    Dump all registers using log callback (write one line).
    """
    width = max(len(name) for name, type in regs._fields_)
    name_format = "%% %us" % width
    for name, type in regs._fields_:
        value = getattr(regs, name)
        name = name_format % name
        if sizeof(type) == 32:
            value = formatUintHex32(value)
        elif sizeof(type) == 16:
            value = formatUintHex16(value)
        else:
            value = formatWordHex(value)
        log("%s = %s" % (name, value))


def readBits(value, bitmasks):
    """
    Extract bits from the integer value using a list of bit masks.
    bitmasks is a list of tuple (mask, text).

    >>> bitmask = (
    ...    (1, "exec"),
    ...    (2, "write"),
    ...    (4, "read"))
    ...
    >>> readBits(5, bitmask)
    ['exec', 'read']
    >>> readBits(12, bitmask)
    ['read', '8']
    """
    bitset = []
    for mask, item in bitmasks:
        if not value & mask:
            continue
        bitset.append(item)
        value = value & ~mask
    if value:
        bitset.append(str(value))
    return bitset


def formatBits(value, bitmasks, empty_text=None, format_value=str):
    """
    Format a value using a bitmask: see readBits() functions.

    >>> bitmask = (
    ...    (1, "exec"),
    ...    (2, "write"),
    ...    (4, "read"))
    ...
    >>> formatBits(5, bitmask, empty_text="no permission")
    '<exec|read> (5)'
    >>> formatBits(0, bitmask, empty_text="no permission")
    'no permission'
    """

    orig_value = value
    text = readBits(value, bitmasks)
    if text:
        text = "%s" % ("|".join(text))
        if value:
            text = "<%s> (%s)" % (text, format_value(orig_value))
        return text
    else:
        if empty_text:
            return empty_text
        else:
            return str(value)


LOCAL_TIMEZONE_OFFSET = datetime.fromtimestamp(
    0) - datetime.utcfromtimestamp(0)

# Start of UNIX timestamp (Epoch): 1st January 1970 at 00:00
UNIX_TIMESTAMP_T0 = datetime(1970, 1, 1)


def timestampUNIX(value, is_local):
    """
    Convert an UNIX (32-bit) timestamp to datetime object. Timestamp value
    is the number of seconds since the 1st January 1970 at 00:00. Maximum
    value is 2147483647: 19 january 2038 at 03:14:07.

    May raise ValueError for invalid value: value have to be in 0..2147483647.

    >>> timestampUNIX(0, False)
    datetime.datetime(1970, 1, 1, 0, 0)
    >>> timestampUNIX(1154175644.37, False)
    datetime.datetime(2006, 7, 29, 12, 20, 44, 370000)
    """
    timestamp = UNIX_TIMESTAMP_T0 + timedelta(seconds=value)
    if is_local:
        timestamp += LOCAL_TIMEZONE_OFFSET
    return timestamp


def locateProgram(program):
    """
    Locate a program using the PATH environment variable. Return the
    unchanged program value if it's not possible to get the full
    program path.
    """
    if isabs(program):
        return program
    if dirname(program):
        # ./test => $PWD/./test
        # ../python => $PWD/../python
        program = path_join(getcwd(), program)
        program = normpath(program)
        return program
    paths = getenv('PATH')
    if not paths:
        return program
    for path in paths.split(pathsep):
        filename = path_join(path, program)
        if access(filename, X_OK):
            return filename
    return program


def minmax(min_value, value, max_value):
    """
    Restrict value to [min_value; max_value]

    >>> minmax(-2, -3, 10)
    -2
    >>> minmax(-2, 27, 10)
    10
    >>> minmax(-2, 0, 10)
    0
    """
    return min(max(min_value, value), max_value)


def inverseDict(data):
    """
    Inverse a dictionary.

    >>> inverseDict({"0x10": 16, "0x20": 32}) == {32: '0x20', 16: '0x10'}
    True
    """
    result = {}
    for key, value in data.items():
        result[value] = key
    return result


def signal_to_exitcode(signum):
    """
    Converts a signal number to an exit code.
    UNIX: https://www.gnu.org/software/bash/manual/html_node/Exit-Status.html

    >>> import signal
    >>> signal_to_exitcode(signal.SIGQUIT)
    131
    """
    return 128 + signum



from ptrace.syscall.posix_constants import SYSCALL_ARG_DICT

RLIMIT_RESOURCE = {
    0: "RLIMIT_CPU",
    1: "RLIMIT_FSIZE",
    2: "RLIMIT_DATA",
    3: "RLIMIT_STACK",
    4: "RLIMIT_CORE",
    5: "RLIMIT_RSS",
    6: "RLIMIT_MEMLOCK",
    7: "RLIMIT_NPROC",
    8: "RLIMIT_NOFILE",
    9: "RLIMIT_SBSIZE",
    10: "RLIMIT_VMEM",
}

SIGPROCMASK_HOW = {1: " SIG_BLOCK", 2: "SIG_UNBLOCK", 3: "SIG_SETMASK"}

SYSCALL_ARG_DICT.update({
    "getrlimit": {"resource": RLIMIT_RESOURCE},
    "setrlimit": {"resource": RLIMIT_RESOURCE},
    "sigprocmask": {"how": SIGPROCMASK_HOW},
    "rt_sigprocmask": {"how": SIGPROCMASK_HOW},
})



from ptrace.syscall.socketcall_constants import (
    SOCKET_FAMILY, SOCKET_PROTOCOL,
    SETSOCKOPT_LEVEL, SETSOCKOPT_OPTNAME)

SYSCALL_ARG_DICT = {
    "lseek": {
        "origin": {0: "SEEK_SET", 1: "SEEK_CUR", 2: "SEEK_END"},
    },
    "futex": {
        "op": {
            0: "FUTEX_WAIT",
            1: "FUTEX_WAKE",
            2: "FUTEX_FD",
            3: "FUTEX_REQUEUE",
            4: "FUTEX_CMP_REQUEUE",
            5: "FUTEX_WAKE_OP",
            6: "FUTEX_LOCK_PI",
            7: "FUTEX_UNLOCK_PI",
            8: "FUTEX_TRYLOCK_PI",
        },
    },
    "fcntl": {
        "cmd": {
            0: "F_DUPFD",
            1: "F_GETFD",
            2: "F_SETFD",
            3: "F_GETFL",
            4: "F_SETFL",
            5: "F_GETOWN",
            6: "F_SETOWN",
            7: "F_GETLK",
            8: "F_SETLK",
            9: "F_SETLKW",
        },
    },
    "ipc": {
        "call": {
            1: "SEMOP",
            2: "SEMGET",
            3: "SEMCTL",
            4: "SEMTIMEDOP",
            11: "MSGSND",
            12: "MSGRCV",
            13: "MSGGET",
            14: "MSGCTL",
            21: "SHMAT",
            22: "SHMDT",
            23: "SHMGET",
            24: "SHMCTL",
        },
    },
    "socket": {
        "domain": SOCKET_FAMILY,
        "protocol": SOCKET_PROTOCOL,
    },
    "getsockopt": {
        "level": SETSOCKOPT_LEVEL,
        "optname": SETSOCKOPT_OPTNAME,
    },
}

SYSCALL_ARG_DICT["setsockopt"] = SYSCALL_ARG_DICT["getsockopt"]



"""
Constants about the operating system:

 - RUNNING_PYPY (bool)
 - RUNNING_LINUX (bool)
 - RUNNING_FREEBSD (bool)
 - RUNNING_OPENBSD (bool)
 - RUNNING_MACOSX (bool)
 - RUNNING_BSD (bool)
 - HAS_PROC (bool)
 - HAS_PTRACE (bool)
"""

from sys import platform, version, version_info

RUNNING_PYTHON3 = version_info[0] == 3
RUNNING_PYPY = ("pypy" in version.lower())
RUNNING_LINUX = platform.startswith('linux')
RUNNING_FREEBSD = (platform.startswith('freebsd')
                   or platform.startswith('gnukfreebsd'))
RUNNING_OPENBSD = platform.startswith('openbsd')
RUNNING_MACOSX = (platform == 'darwin')
RUNNING_BSD = RUNNING_FREEBSD or RUNNING_MACOSX or RUNNING_OPENBSD

HAS_PROC = RUNNING_LINUX
HAS_PTRACE = (RUNNING_BSD or RUNNING_LINUX)



from os import strerror
from errno import errorcode

from ptrace.cpu_info import CPU_X86_64, CPU_POWERPC, CPU_I386, CPU_ARM32, CPU_AARCH64
from ptrace.ctypes_tools import ulong2long, formatAddress, formatWordHex
from ptrace.func_call import FunctionCall
from ptrace.syscall import SYSCALL_NAMES, SYSCALL_PROTOTYPES, SyscallArgument
from ptrace.syscall.socketcall import setupSocketCall
from ptrace.os_tools import RUNNING_LINUX, RUNNING_BSD
from ptrace.cpu_info import CPU_WORD_SIZE
from ptrace.binding.cpu import CPU_INSTR_POINTER

if CPU_POWERPC:
    SYSCALL_REGISTER = "gpr0"
elif CPU_ARM32:
    SYSCALL_REGISTER = "r7"
elif CPU_AARCH64:
    SYSCALL_REGISTER = "r8"
elif RUNNING_LINUX:
    if CPU_X86_64:
        SYSCALL_REGISTER = "orig_rax"
    else:
        SYSCALL_REGISTER = "orig_eax"
else:
    if CPU_X86_64:
        SYSCALL_REGISTER = "rax"
    else:
        SYSCALL_REGISTER = "eax"

if CPU_ARM32:
    RETURN_VALUE_REGISTER = "r0"
elif CPU_AARCH64:
    RETURN_VALUE_REGISTER = "r0"
elif CPU_I386:
    RETURN_VALUE_REGISTER = "eax"
elif CPU_X86_64:
    RETURN_VALUE_REGISTER = "rax"
elif CPU_POWERPC:
    RETURN_VALUE_REGISTER = "result"
else:
    raise NotImplementedError("Unsupported CPU architecture")

PREFORMAT_ARGUMENTS = {
    "select": (1, 2, 3),
    "execve": (0, 1, 2),
    "clone": (0, 1),
}


class PtraceSyscall(FunctionCall):

    def __init__(self, process, options, regs=None):
        FunctionCall.__init__(self, "syscall", options, SyscallArgument)
        self.process = process
        self.restype = "long"
        self.result = None
        self.result_text = None
        self.instr_pointer = None
        if not regs:
            regs = self.process.getregs()
        self.readSyscall(regs)

    def enter(self, regs=None):
        if not regs:
            regs = self.process.getregs()
        argument_values = self.readArgumentValues(regs)
        self.readArguments(argument_values)

        if self.name == "socketcall" and self.options.replace_socketcall:
            setupSocketCall(self, self.process, self[0], self[1].value)

        # Some arguments are lost after the syscall, so format them now
        if self.name in PREFORMAT_ARGUMENTS:
            for index in PREFORMAT_ARGUMENTS[self.name]:
                argument = self.arguments[index]
                argument.format()

        if self.options.instr_pointer:
            self.instr_pointer = getattr(regs, CPU_INSTR_POINTER)

    def readSyscall(self, regs):
        # Read syscall number
        self.syscall = getattr(regs, SYSCALL_REGISTER)
        # Get syscall variables
        self.name = SYSCALL_NAMES.get(
            self.syscall, "syscall<%s>" % self.syscall)

    def readArgumentValues(self, regs):
        if CPU_X86_64:
            return (regs.rdi, regs.rsi, regs.rdx, regs.r10, regs.r8, regs.r9)
        if CPU_ARM32:
            return (regs.r0, regs.r1, regs.r2, regs.r3, regs.r4, regs.r5, regs.r6)
        if CPU_AARCH64:
            return (regs.r0, regs.r1, regs.r2, regs.r3, regs.r4, regs.r5, regs.r6, regs.r7)
        if RUNNING_BSD:
            sp = self.process.getStackPointer()
            return [self.process.readWord(sp + index * CPU_WORD_SIZE)
                    for index in range(1, 6 + 1)]
        if CPU_I386:
            return (regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi, regs.ebp)
        if CPU_POWERPC:
            return (regs.gpr3, regs.gpr4, regs.gpr5, regs.gpr6, regs.gpr7, regs.gpr8)
        raise NotImplementedError()

    def readArguments(self, argument_values):
        if self.name in SYSCALL_PROTOTYPES:
            self.restype, formats = SYSCALL_PROTOTYPES[self.name]
            for value, format in zip(argument_values, formats):
                argtype, argname = format
                self.addArgument(value=value, name=argname, type=argtype)
        else:
            for value in argument_values:
                self.addArgument(value=value)

    def exit(self):
        if self.name in PREFORMAT_ARGUMENTS:
            preformat = set(PREFORMAT_ARGUMENTS[self.name])
        else:
            preformat = set()

        # Data pointed by arguments may have changed during the syscall
        # e.g. uname() syscall
        for index, argument in enumerate(self.arguments):
            if index in preformat:
                # Don't lose preformatted arguments
                continue
            if argument.type and not argument.type.endswith("*"):
                continue
            argument.text = None

        self.result = self.process.getreg(RETURN_VALUE_REGISTER)

        if self.restype.endswith("*"):
            text = formatAddress(self.result)
        else:
            uresult = self.result
            self.result = ulong2long(self.result)
            if self.result < 0 and (-self.result) in errorcode:
                errcode = -self.result
                text = "%s %s (%s)" % (
                    self.result, errorcode[errcode], strerror(errcode))
            elif not(0 <= self.result <= 9):
                text = "%s (%s)" % (self.result, formatWordHex(uresult))
            else:
                text = str(self.result)
        self.result_text = text
        return text

    def __str__(self):
        return "<Syscall name=%r>" % self.name



from logging import error
from ptrace.disasm import HAS_DISASSEMBLER
from signal import SIGFPE, SIGSEGV, SIGABRT
try:
    from signal import SIGCHLD
except ImportError:
    SIGCHLD = None
try:
    from signal import SIGBUS
except ImportError:
    SIGBUS = None
from ptrace.os_tools import RUNNING_LINUX
from ptrace.cpu_info import CPU_64BITS
from ptrace.debugger import ProcessEvent
from ptrace.error import PtraceError
from ptrace import signalName
from ptrace.debugger.signal_reason import (
    DivisionByZero, Abort, StackOverflow,
    InvalidMemoryAccess, InvalidRead, InvalidWrite,
    InstructionError, ChildExit)
from ptrace.debugger.parse_expr import parseExpression
import re

# Match a pointer dereference (e.g. "DWORD [EDX+0x8]")
DEREF_REGEX = r'(?P<deref_size>(BYTE|WORD|DWORD|DQWORD) )?\[(?P<deref>[^]]+)\]'

NAMED_WORD_SIZE = {
    'BYTE': 1,
    'WORD': 2,
    'DWORD': 4,
    'DQWORD': 8,
}

# Match any Intel instruction (e.g. "ADD")
INSTR_REGEX = '(?:[A-Z]{3,10})'


def findDerefSize(match):
    name = match.group("deref_size")
    try:
        return NAMED_WORD_SIZE[name]
    except KeyError:
        return None


def evalFaultAddress(process, match):
    expr = match.group('deref')
    if not expr:
        return None
    try:
        return parseExpression(process, expr)
    except ValueError as err:
        error("err: %s" % err)
        return None


class ProcessSignal(ProcessEvent):

    def __init__(self, signum, process):
        # Initialize attributes
        self.name = signalName(signum)
        ProcessEvent.__init__(self, process, "Signal %s" % self.name)
        self.signum = signum
        self.reason = None

    def _analyze(self):
        if self.signum in (SIGSEGV, SIGBUS):
            self.memoryFault()
        elif self.signum == SIGFPE:
            self.mathError()
        elif self.signum == SIGCHLD:
            self.childExit()
        elif self.signum == SIGABRT:
            self.reason = Abort()
        return self.reason

    def getInstruction(self):
        if not HAS_DISASSEMBLER:
            return None
        try:
            return self.process.disassembleOne()
        except PtraceError:
            return None

    def memoryFaultInstr(self, instr, fault_address):
        asm = instr.text

        # Invalid write (e.g. "MOV [...], value")
        match = re.search(r"^(?:MOV|TEST)[A-Z]* %s," % DEREF_REGEX, asm)
        if match:
            if fault_address is None:
                fault_address = evalFaultAddress(self.process, match)
            self.reason = InvalidWrite(fault_address, size=findDerefSize(match),
                                       instr=instr, process=self.process)
            return

        # Invalid read (e.g. "CMP BYTE [EAX+EDX-0x1], 0x0")
        match = re.search(r"^%s %s," % (INSTR_REGEX, DEREF_REGEX), asm)
        if match:
            if fault_address is None:
                fault_address = evalFaultAddress(self.process, match)
            self.reason = InvalidRead(fault_address, size=findDerefSize(match),
                                      instr=instr, process=self.process)
            return

        # Invalid read (e.g. "MOV reg, [...]")
        match = re.match(r"%s [^,]+, %s" % (INSTR_REGEX, DEREF_REGEX), asm)
        if match:
            if fault_address is None:
                fault_address = evalFaultAddress(self.process, match)
            self.reason = InvalidRead(fault_address, size=findDerefSize(match),
                                      instr=instr, process=self.process)
            return

        # MOVS* and SCAS* instructions (e.g. "MOVSB" or "REP SCASD")
        match = re.search(
            r"^(?:REP(?:NZ)? )?(?P<operator>MOVS|SCAS)(?P<suffix>[BWD])?", asm)
        if match:
            self.reason = self.movsInstr(fault_address, instr, match)
            return

    def movsInstr(self, fault_address, instr, match):
        operator = match.group("operator")
        suffix = match.group("suffix")
        size = {'B': 1, 'W': 2, 'D': 4}.get(suffix)
        error_cls = InvalidMemoryAccess
        try:
            process = self.process
            if CPU_64BITS:
                source_reg = 'rsi'
                dest_reg = 'rdi'
            else:
                source_reg = 'esi'
                dest_reg = 'edi'
            source_addr = process.getreg(source_reg)
            registers = {source_reg: source_addr}
            write = (operator == 'MOVS')
            if write:
                dest_addr = process.getreg(dest_reg)
                registers[dest_reg] = dest_addr

            if fault_address is not None:
                if fault_address == source_addr:
                    error_cls = InvalidRead
                if write and fault_address == dest_addr:
                    error_cls = InvalidWrite
            else:
                if write:
                    fault_address = (source_addr, dest_addr)
                else:
                    fault_address = (source_addr,)
        except PtraceError:
            registers = {}
        return error_cls(fault_address, size=size, instr=instr,
                         registers=registers, process=self.process)

    def getSignalInfo(self):
        if RUNNING_LINUX:
            return self.process.getsiginfo()
        else:
            return None

    def memoryFault(self):
        # Get fault
        siginfo = self.getSignalInfo()
        if siginfo:
            fault_address = siginfo._sigfault._addr
            if not fault_address:
                fault_address = 0
        else:
            fault_address = None

        # Get current instruction
        instr = self.getInstruction()

        # Call to invalid address?
        if fault_address is not None:
            try:
                ip = self.process.getInstrPointer()
                if ip == fault_address:
                    self.reason = InstructionError(ip, process=self.process)
                    return
            except PtraceError:
                pass

        # Stack overflow?
        stack = self.process.findStack()
        if stack:
            sp = self.process.getStackPointer()
            if not (stack.start <= sp <= stack.end):
                self.reason = StackOverflow(
                    sp, stack, instr=instr, process=self.process)
                return

        # Guess error type using the assembler instruction
        if instr:
            self.memoryFaultInstr(instr, fault_address)
            if self.reason:
                return

        # Last chance: use generic invalid memory access error
        self.reason = InvalidMemoryAccess(
            fault_address, instr=instr, process=self.process)

    def mathError(self):
        instr = self.getInstruction()
        if not instr:
            return
        match = re.match(r"I?DIV (.*)", instr.text)
        if not match:
            return
        self.reason = DivisionByZero(instr=instr, process=self.process)

    def childExit(self):
        siginfo = self.getSignalInfo()
        if siginfo:
            child = siginfo._sigchld
            self.reason = ChildExit(child.pid, child.status, child.uid)
        else:
            self.reason = ChildExit()

    def display(self, log=None):
        self._analyze()
        if not log:
            log = error
        log("-" * 60)
        log("PID: %s" % self.process.pid)
        log("Signal: %s" % self.name)
        if self.reason:
            self.reason.display(log)
        log("-" * 60)



from ptrace.syscall.names import SYSCALL_NAMES, SOCKET_SYSCALL_NAMES   # noqa
from ptrace.syscall.prototypes import SYSCALL_PROTOTYPES, FILENAME_ARGUMENTS   # noqa
from ptrace.syscall.syscall_argument import SyscallArgument   # noqa
from ptrace.syscall.ptrace_syscall import PtraceSyscall, SYSCALL_REGISTER, RETURN_VALUE_REGISTER   # noqa



from ptrace.debugger.breakpoint import Breakpoint   # noqa
from ptrace.debugger.process_event import (ProcessEvent, ProcessExit,   # noqa
                                           NewProcessEvent, ProcessExecution)
from ptrace.debugger.ptrace_signal import ProcessSignal   # noqa
from ptrace.debugger.process_error import ProcessError   # noqa
from ptrace.debugger.child import ChildError   # noqa
from ptrace.debugger.process import PtraceProcess   # noqa
from ptrace.debugger.debugger import PtraceDebugger, DebuggerError   # noqa
from ptrace.debugger.application import Application   # noqa



from ptrace.ctypes_tools import formatAddress, formatWordHex
from ptrace.error import PtraceError
from ptrace.cpu_info import CPU_I386, CPU_X86_64
from ptrace.process_tools import formatProcessStatus
import re

# Find all Intel registers (in upper case)
if CPU_I386:
    regex = 'E[A-Z]{2}|[CDEFGS]S|[ABCD]L'
elif CPU_X86_64:
    regex = '[ER][A-Z]{2}|[CDEFGS]S|[ABCD]L'
else:
    regex = None
if regex:
    REGISTER_REGEX = re.compile(r'\b(?:%s)\b' % regex)
else:
    REGISTER_REGEX = None


def extractRegisters(process, instr):
    registers = {}
    if not process or not instr or not REGISTER_REGEX:
        return registers
    asm = instr.text
    asm = asm.upper()
    # Skip the operator ("MOV CL, [EAX]" => "CL, [EAX]")
    asm = asm.split(" ", 1)[1]
    for match in REGISTER_REGEX.finditer(asm):
        name = match.group(0)
        name = name.lower()
        try:
            value = process.getreg(name)
            registers[name] = value
        except PtraceError:
            pass
    return registers


def findMappings(addresses, process, size):
    mappings = []
    if addresses is None or not process:
        return mappings
    if not isinstance(addresses, (list, tuple)):
        addresses = (addresses,)
    if not size:
        size = 0
    process_mappings = process.readMappings()
    if not process_mappings:
        return mappings
    for address in addresses:
        address_str = formatAddress(address)
        if 1 < size:
            address_str += "..%s" % formatAddress(address + size - 1)
        found = False
        for map in process_mappings:
            if (map.start <= address < map.end) \
                    or (map.start <= (address + size - 1) < map.end):
                found = True
                mappings.append("%s is part of %s" % (address_str, map))
        if not found:
            mappings.append("%s is not mapped in memory" % address_str)
    return mappings


class SignalInfo(Exception):

    def __init__(self, name, text,
                 address=None, size=None, instr=None,
                 process=None, registers=None):
        Exception.__init__(self, text)
        self.name = name
        self.text = text
        self.instr = instr
        self.registers = extractRegisters(process, instr)
        if registers:
            self.registers.update(registers)
        self.mappings = findMappings(address, process, size)

    def display(self, log):
        log(self.text)
        self.displayExtra(log)
        if self.instr:
            log("- instruction: %s" % self.instr)
        for mapping in self.mappings:
            log("- mapping: %s" % mapping)
        for name, value in self.registers.items():
            log("- register %s=%s" % (name, formatWordHex(value)))

    def displayExtra(self, log):
        pass


class DivisionByZero(SignalInfo):

    def __init__(self, instr=None, process=None):
        SignalInfo.__init__(self, "div_by_zero",
                            "Division by zero", instr=instr, process=process)


class Abort(SignalInfo):

    def __init__(self):
        SignalInfo.__init__(self, "abort",
                            "Program received signal SIGABRT, Aborted.")


class StackOverflow(SignalInfo):

    def __init__(self, stack_ptr, stack_map, instr=None, process=None):
        text = "STACK OVERFLOW! Stack pointer is in %s" % stack_map
        SignalInfo.__init__(self, "stack_overflow", text,
                            address=stack_ptr, registers={
                                '<stack ptr>': stack_ptr},
                            instr=instr, process=process)
        self.stack_ptr = stack_ptr
        self.stack_map = stack_map


class InvalidMemoryAccess(SignalInfo):
    NAME = "invalid_mem_access"
    PREFIX = "Invalid memory access"
    PREFIX_ADDR = "Invalid memory access to %s"

    def __init__(self, address=None, size=None, instr=None, registers=None, process=None):
        """
        address is an integer or a list of integer
        """
        if address is not None:
            if isinstance(address, (list, tuple)):
                arguments = " or ".join(formatAddress(addr)
                                        for addr in address)
            else:
                arguments = formatAddress(address)
            message = self.PREFIX_ADDR % arguments
        else:
            message = self.PREFIX
        if size:
            message += " (size=%s bytes)" % size
        name = self.NAME
        if address is not None:
            name += "-" + formatAddress(address).lower()
        SignalInfo.__init__(self, name, message,
                            address=address, size=size, instr=instr,
                            process=process, registers=registers)


class InvalidRead(InvalidMemoryAccess):
    NAME = "invalid_read"
    PREFIX = "Invalid read"
    PREFIX_ADDR = "Invalid read from %s"


class InvalidWrite(InvalidMemoryAccess):
    NAME = "invalid_write"
    PREFIX = "Invalid write"
    PREFIX_ADDR = "Invalid write to %s"


class InstructionError(SignalInfo):

    def __init__(self, address, process=None):
        SignalInfo.__init__(self, "instr_error",
                            "UNABLE TO EXECUTE CODE AT %s (SEGMENTATION FAULT)" % formatAddress(
                                address),
                            address=address,
                            process=process,
                            registers={'<instr pointer>': address})


class ChildExit(SignalInfo):

    def __init__(self, pid=None, status=None, uid=None):
        if pid is not None and status is not None:
            message = formatProcessStatus(status, "Child process %s" % pid)
        else:
            message = "Child process exited"
        SignalInfo.__init__(self, "child_exit", message)
        self.pid = pid
        self.status = status
        self.uid = uid

    def displayExtra(self, log):
        if self.uid is not None:
            log("Signal sent by user %s" % self.uid)



from sys import exc_info
from traceback import format_exception
from logging import ERROR, getLogger
from ptrace.logging_tools import getLogFunc, changeLogLevel

PTRACE_ERRORS = Exception


def writeBacktrace(logger, log_level=ERROR):
    """
    Write a backtrace into the logger with the specified log level.
    """
    log_func = getLogFunc(logger, log_level)
    try:
        info = exc_info()
        trace = format_exception(*info)
        if trace[0] != "None\n":
            trace = ''.join(trace).rstrip()
            for line in trace.split("\n"):
                log_func(line.rstrip())
            return
    except Exception:
        pass
    log_func("Unable to get backtrace")


def formatError(error):
    """
    Format an error as a string. Write the error type as prefix.
    Eg. "[ValueError] invalid value".
    """
    return "[%s] %s" % (error.__class__.__name__, error)


def writeError(logger, error, title="ERROR", log_level=ERROR):
    """
    Write an error into the logger:
     - logger: the logger (if None, use getLogger())
     - error: the exception object
     - title: error message prefix (e.g. title="Initialization error")
     - log_level: log level of the error

    If the exception is a SystemExit or a KeyboardInterrupt, re-emit
    (raise) the exception and don't write it.
    """
    if not logger:
        logger = getLogger()
    if error.__class__ in (SystemExit, KeyboardInterrupt):
        raise error
    log_func = getLogFunc(logger, log_level)
    log_func("%s: %s" % (title, formatError(error)))
    writeBacktrace(logger, log_level=changeLogLevel(log_level, -1))


class PtraceError(Exception):
    """
    Ptrace error: have the optional attributes errno and pid.
    """

    def __init__(self, message, errno=None, pid=None):
        Exception.__init__(self, message)
        self.errno = errno
        self.pid = pid



from ctypes import Structure, c_int, c_uint, c_ulong, c_void_p, c_char

PIOD_READ_D = 1
PIOD_WRITE_D = 2
PIOD_READ_I = 3
PIOD_WRITE_I = 4

size_t = c_ulong
pid_t = c_int

# /usr/include/machine/reg.h


class reg(Structure):
    _fields_ = (
        ("eax", c_uint),
        ("ecx", c_uint),
        ("edx", c_uint),
        ("ebx", c_uint),
        ("esp", c_uint),
        ("ebp", c_uint),
        ("esi", c_uint),
        ("edi", c_uint),
        ("eip", c_uint),
        ("eflags", c_uint),
        ("cs", c_uint),
        ("ss", c_uint),
        ("ds", c_uint),
        ("es", c_uint),
        ("fs", c_uint),
        ("gs", c_uint),
    )


class fpreg(Structure):
    _fields_ = (
        ("__data", c_char * 116),
    )


class ptrace_io_desc(Structure):
    _fields_ = (
        ("piod_op", c_int),
        ("piod_offs", c_void_p),
        ("piod_addr", c_void_p),
        ("piod_len", size_t),
    )



#!/usr/bin/python
from ptrace import PtraceError
from ptrace.debugger import (PtraceDebugger, Application,
                             ProcessExit, ProcessSignal, NewProcessEvent, ProcessExecution)
from ptrace.syscall import (SYSCALL_NAMES, SYSCALL_PROTOTYPES,
                            FILENAME_ARGUMENTS, SOCKET_SYSCALL_NAMES)
from ptrace.func_call import FunctionCallOptions
from sys import stderr, exit
from optparse import OptionParser
from logging import getLogger, error
from ptrace.error import PTRACE_ERRORS, writeError
from ptrace.ctypes_tools import formatAddress
from ptrace.tools import signal_to_exitcode
import sys
import re


class SyscallTracer(Application):

    def __init__(self):
        Application.__init__(self)

        # Parse self.options
        self.parseOptions()

        # Setup output (log)
        self.setupLog()

    def setupLog(self):
        if self.options.output:
            fd = open(self.options.output, 'w')
            self._output = fd
        else:
            fd = stderr
            self._output = None
        self._setupLog(fd)

    def parseOptions(self):
        parser = OptionParser(
            usage="%prog [options] -- program [arg1 arg2 ...]")
        self.createCommonOptions(parser)
        parser.add_option("--enter", help="Show system call enter and exit",
                          action="store_true", default=False)
        parser.add_option("--profiler", help="Use profiler",
                          action="store_true", default=False)
        parser.add_option("--type", help="Display arguments type and result type (default: no)",
                          action="store_true", default=False)
        parser.add_option("--name", help="Display argument name (default: no)",
                          action="store_true", default=False)
        parser.add_option("--string-length", "-s", help="String max length (default: 300)",
                          type="int", default=300)
        parser.add_option("--array-count", help="Maximum number of array items (default: 20)",
                          type="int", default=20)
        parser.add_option("--raw-socketcall", help="Raw socketcall form",
                          action="store_true", default=False)
        parser.add_option("--output", "-o", help="Write output to specified log file",
                          type="str")
        parser.add_option("--ignore-regex", help="Regex used to filter syscall names (e.g. --ignore='^(gettimeofday|futex|f?stat)')",
                          type="str")
        parser.add_option("--address", help="Display structure address",
                          action="store_true", default=False)
        parser.add_option("--syscalls", '-e', help="Comma separated list of shown system calls (other will be skipped)",
                          type="str", default=None)
        parser.add_option("--socket", help="Show only socket functions",
                          action="store_true", default=False)
        parser.add_option("--filename", help="Show only syscall using filename",
                          action="store_true", default=False)
        parser.add_option("--show-pid",
                          help="Prefix line with process identifier",
                          action="store_true", default=False)
        parser.add_option("--list-syscalls",
                          help="Display system calls and exit",
                          action="store_true", default=False)
        parser.add_option("-i", "--show-ip",
                          help="print instruction pointer at time of syscall",
                          action="store_true", default=False)

        self.createLogOptions(parser)

        self.options, self.program = parser.parse_args()

        if self.options.list_syscalls:
            syscalls = list(SYSCALL_NAMES.items())
            syscalls.sort(key=lambda data: data[0])
            for num, name in syscalls:
                print("% 3s: %s" % (num, name))
            exit(0)

        if self.options.pid is None and not self.program:
            parser.print_help()
            exit(1)

        # Create "only" filter
        only = set()
        if self.options.syscalls:
            # split by "," and remove spaces
            for item in self.options.syscalls.split(","):
                item = item.strip()
                if not item or item in only:
                    continue
                ok = True
                valid_names = list(SYSCALL_NAMES.values())
                for name in only:
                    if name not in valid_names:
                        print("ERROR: unknown syscall %r" % name, file=stderr)
                        ok = False
                if not ok:
                    print(file=stderr)
                    print(
                        "Use --list-syscalls options to get system calls list", file=stderr)
                    exit(1)
                # remove duplicates
                only.add(item)
        if self.options.filename:
            for syscall, format in SYSCALL_PROTOTYPES.items():
                restype, arguments = format
                if any(argname in FILENAME_ARGUMENTS for argtype, argname in arguments):
                    only.add(syscall)
        if self.options.socket:
            only |= SOCKET_SYSCALL_NAMES
        self.only = only
        if self.options.ignore_regex:
            try:
                self.ignore_regex = re.compile(self.options.ignore_regex)
            except Exception as err:
                print("Invalid regular expression! %s" % err)
                print("(regex: %r)" % self.options.ignore_regex)
                exit(1)
        else:
            self.ignore_regex = None

        if self.options.fork:
            self.options.show_pid = True

        self.processOptions()

    def ignoreSyscall(self, syscall):
        name = syscall.name
        if self.only and (name not in self.only):
            return True
        if self.ignore_regex and self.ignore_regex.match(name):
            return True
        return False

    def displaySyscall(self, syscall):
        text = syscall.format()
        if syscall.result is not None:
            text = "%-40s = %s" % (text, syscall.result_text)
        prefix = []
        if self.options.show_pid:
            prefix.append("[%s]" % syscall.process.pid)
        if self.options.show_ip:
            prefix.append("[%s]" % formatAddress(syscall.instr_pointer))
        if prefix:
            text = ''.join(prefix) + ' ' + text
        error(text)

    def syscallTrace(self, process):
        # First query to break at next syscall
        self.prepareProcess(process)
        exitcode = 0
        while True:
            # No more process? Exit
            if not self.debugger:
                break

            # Wait until next syscall enter
            try:
                event = self.debugger.waitSyscall()
            except ProcessExit as event:
                self.processExited(event)
                if event.exitcode is not None:
                    exitcode = event.exitcode
                continue
            except ProcessSignal as event:
                event.display()
                event.process.syscall(event.signum)
                exitcode = signal_to_exitcode(event.signum)
                continue
            except NewProcessEvent as event:
                self.newProcess(event)
                continue
            except ProcessExecution as event:
                self.processExecution(event)
                continue

            # Process syscall enter or exit
            self.syscall(event.process)
        return exitcode

    def syscall(self, process):
        state = process.syscall_state
        syscall = state.event(self.syscall_options)
        if syscall and (syscall.result is not None or self.options.enter):
            self.displaySyscall(syscall)

        # Break at next syscall
        process.syscall()

    def processExited(self, event):
        # Display syscall which has not exited
        state = event.process.syscall_state
        if (state.next_event == "exit") \
                and (not self.options.enter) \
                and state.syscall:
            self.displaySyscall(state.syscall)

        # Display exit message
        error("*** %s ***" % event)

    def prepareProcess(self, process):
        process.syscall()
        process.syscall_state.ignore_callback = self.ignoreSyscall

    def newProcess(self, event):
        process = event.process
        error("*** New process %s ***" % process.pid)
        self.prepareProcess(process)
        process.parent.syscall()

    def processExecution(self, event):
        process = event.process
        error("*** Process %s execution ***" % process.pid)
        process.syscall()

    def runDebugger(self):
        # Create debugger and traced process
        self.setupDebugger()
        process = self.createProcess()
        if not process:
            return

        self.syscall_options = FunctionCallOptions(
            write_types=self.options.type,
            write_argname=self.options.name,
            string_max_length=self.options.string_length,
            replace_socketcall=not self.options.raw_socketcall,
            write_address=self.options.address,
            max_array_count=self.options.array_count,
        )
        self.syscall_options.instr_pointer = self.options.show_ip

        return self.syscallTrace(process)

    def main(self):
        if self.options.profiler:
            from ptrace.profiler import runProfiler
            exitcode = runProfiler(getLogger(), self._main)
        else:
            exitcode = self._main()
        if self._output is not None:
            self._output.close()
        sys.exit(exitcode)

    def _main(self):
        self.debugger = PtraceDebugger()
        exitcode = 0
        try:
            exitcode = self.runDebugger()
        except ProcessExit as event:
            self.processExited(event)
            if event.exitcode is not None:
                exitcode = event.exitcode
        except PtraceError as err:
            error("ptrace() error: %s" % err)
            if err.errno is not None:
                exitcode = err.errno
        except KeyboardInterrupt:
            error("Interrupted.")
            exitcode = 1
        except PTRACE_ERRORS as err:
            writeError(getLogger(), err, "Debugger error")
            exitcode = 1
        self.debugger.quit()
        return exitcode

    def createChild(self, program):
        pid = Application.createChild(self, program)
        error("execve(%s, %s, [/* 40 vars */]) = %s" % (
            program[0], program, pid))
        return pid


if __name__ == "__main__":
    SyscallTracer().main()



from ptrace.func_arg import FunctionArgument


class FunctionCallOptions(object):
    """
    Options to format a function call and its arguments.
    """

    def __init__(self,
                 write_types=False, write_argname=False,
                 replace_socketcall=True, string_max_length=300,
                 write_address=False, max_array_count=20):
        self.write_address = write_address
        self.write_types = write_types
        self.write_argname = write_argname
        self.replace_socketcall = replace_socketcall
        self.string_max_length = string_max_length
        self.max_array_count = max_array_count
        self.instr_pointer = False


class FunctionCall(object):
    """
    A function call. Attributes:
     - name (str): function name
     - arguments: list of FunctionArgument objects
     - restype (str, optional): result type
     - resvalue (optional): result value
     - argument_class: class used to build the new arguments

    Methods:
     - format(): create a string representation of the call
     - addArgument(): add a new argument
     - clearArguments(): remove all arguments
    """

    def __init__(self, name, options, argument_class=FunctionArgument):
        self.name = name
        self.options = options
        self.arguments = []
        self.restype = None
        self.resvalue = None
        self.argument_class = argument_class

    def addArgument(self, value=None, name=None, type=None):
        arg = self.argument_class(
            self, len(self.arguments), self.options, value, type, name)
        self.arguments.append(arg)

    def clearArguments(self):
        self.arguments = []

    def __getitem__(self, key):
        if isinstance(key, str):
            for arg in self.arguments:
                if arg.name == key:
                    return arg
            raise KeyError("%r has no argument called %r" % (self, key))
        else:
            # Integer key
            return self.arguments[key]

    def format(self):
        arguments = [arg.format() for arg in self.arguments]

        # Remove empty optionnal arguments
        while arguments and not arguments[-1]:
            arguments.pop(-1)

        arguments = ", ".join(arguments)
        if self.restype and self.options.write_types:
            return "%s %s(%s)" % (self.restype, self.name, arguments)
        else:
            return "%s(%s)" % (self.name, arguments)

    def __repr__(self):
        return "<FunctionCall name=%r>" % self.name



#!/usr/bin/env python
from ptrace.debugger.debugger import PtraceDebugger
from ptrace.debugger.child import createChild
from ptrace.tools import locateProgram
from sys import stderr, argv, exit

def playWithProcess(process):
    # Do anything you want with the process here...
    print("Dump process registers")
    process.dumpRegs()
    print("Continue process execution")
    process.cont()
    print("Wait next process event...")
    event = process.waitEvent()
    print("New process event: %s" % event)

def traceProgram(arguments):
    # Copy the environment variables
    env = None

    # Get the full path of the program
    arguments[0] = locateProgram(arguments[0])

    # Create the child process
    return createChild(arguments, False, env)

def main():
    # Check the command line
    if len(argv) < 2:
        print("usage: %s program [arg1 arg2 ...]" % argv[0], file=stderr)
        print("   or: %s pid" % argv[0], file=stderr)
        exit(1)

    # Get the process identifier
    is_attached = False
    has_pid = False
    if len(argv) == 2:
        try:
            # User asked to attach a process
            pid = int(argv[1])
            has_pid = True
        except ValueError:
            pass

    if not has_pid:
        # User asked to create a new program and trace it
        arguments = argv[1:]
        pid = traceProgram(arguments)
        is_attached = True

    # Create the debugger and attach the process
    dbg = PtraceDebugger()
    process = dbg.addProcess(pid, is_attached)

    # Play with the process and then quit
    playWithProcess(process)
    dbg.quit()

if __name__ == "__main__":
    main()




from ptrace.tools import minmax
from logging import ERROR, WARNING, INFO, DEBUG


def getLogFunc(logger, level):
    """
    Get the logger function for the specified logging level.
    """
    if level == ERROR:
        return logger.error
    elif level == WARNING:
        return logger.warning
    elif level == INFO:
        return logger.info
    elif level == DEBUG:
        return logger.debug
    else:
        return logger.error


def changeLogLevel(level, delta):
    """
    Compute log level and make sure that the result is in DEBUG..ERROR.

    >>> changeLogLevel(ERROR, -1) == WARNING
    True
    >>> changeLogLevel(DEBUG, 1) == INFO
    True
    """
    return minmax(DEBUG, level + delta * 10, ERROR)



"""
:[diStorm64 1.7.27}:
Copyright RageStorm (C) 2007, Gil Dabah

diStorm is licensed under the BSD license.
http://ragestorm.net/distorm/
---
Python binding of diStorm64 library written by Victor Stinner
"""

from ctypes import cdll, c_long, c_ulong, c_int, c_uint, c_char, POINTER, Structure, addressof, byref, c_void_p, create_string_buffer, sizeof, cast

# Define (u)int32_t and (u)int64_t types
int32_t = c_int
uint32_t = c_uint
if sizeof(c_ulong) == 8:
    int64_t = c_long
    uint64_t = c_ulong
else:
    from ctypes import c_longlong, c_ulonglong
    assert sizeof(c_longlong) == 8
    assert sizeof(c_ulonglong) == 8
    int64_t = c_longlong
    uint64_t = c_ulonglong

SUPPORT_64BIT_OFFSET = True
if SUPPORT_64BIT_OFFSET:
    _OffsetType = uint64_t
else:
    _OffsetType = uint32_t

LIB_FILENAME = 'libdistorm64.so'
distorm = cdll.LoadLibrary(LIB_FILENAME)
Decode16Bits = 0
Decode32Bits = 1
Decode64Bits = 2
DECODERS = (Decode16Bits, Decode32Bits, Decode64Bits)

internal_decode = distorm.internal_decode

DECRES_NONE = 0
DECRES_SUCCESS = 1
DECRES_MEMORYERR = 2
DECRES_INPUTERR = 3

MAX_INSTRUCTIONS = 100
MAX_TEXT_SIZE = 60


class _WString(Structure):
    _fields_ = (
        ("pos", c_uint),
        ("p", c_char * MAX_TEXT_SIZE),
    )

    def __str__(self):
        # FIXME: Use pos?
        return self.p


class _DecodedInst(Structure):
    _fields_ = (
        ("mnemonic", _WString),
        ("operands", _WString),
        ("instructionHex", _WString),
        ("size", c_uint),
        ("offset", _OffsetType),
    )

    def __str__(self):
        return "%s %s" % (self.mnemonic, self.operands)


internal_decode.argtypes = (_OffsetType, c_void_p,
                            c_int, c_int, c_void_p, c_uint, POINTER(c_uint))


def Decode(codeOffset, code, dt=Decode32Bits):
    """
    Errors: TypeError, IndexError, MemoryError, ValueError
    """
    # Check arguments
    if not isinstance(codeOffset, int):
        raise TypeError("codeOffset have to be an integer")
    if not isinstance(code, bytes):
        raise TypeError("code have to be a bytes, not %s"
                        % (type(code).__name__,))
    if dt not in DECODERS:
        raise IndexError(
            "Decoding-type must be either Decode16Bits, Decode32Bits or Decode64Bits.")

    # Allocate memory for decoder
    code_buffer = create_string_buffer(code)
    decodedInstructionsCount = c_uint()
    result = create_string_buffer(sizeof(_DecodedInst) * MAX_INSTRUCTIONS)

    # Prepare arguments
    codeLen = len(code)
    code = addressof(code_buffer)
    while codeLen:
        # Call internal decoder
        res = internal_decode(codeOffset, code, codeLen, dt, result,
                              MAX_INSTRUCTIONS, byref(decodedInstructionsCount))

        # Check for errors
        if res == DECRES_INPUTERR:
            raise ValueError("Invalid argument")
        count = decodedInstructionsCount.value
        if res == DECRES_MEMORYERR and not count:
            raise MemoryError()

        # No more instruction
        if not count:
            break

        # Yield instruction and compute decoded size
        size = 0
        instr_array = cast(result, POINTER(_DecodedInst))
        for index in range(count):
            instr = instr_array[index]
            size += instr.size
            yield instr

        # Update parameters to move to next instructions
        code += size
        codeOffset += size
        codeLen -= size



# Linux kernel syscall list from Linux 2.6.21 for i386
#
# List extracted from Linux kernel source code, see:
#    arch/i386/kernel/syscall_table.S
SYSCALL_NAMES = {
    0: "restart_syscall",
    1: "exit",
    2: "fork",
    3: "read",
    4: "write",
    5: "open",
    6: "close",
    7: "waitpid",
    8: "creat",
    9: "link",
    10: "unlink",
    11: "execve",
    12: "chdir",
    13: "time",
    14: "mknod",
    15: "chmod",
    16: "lchown16",
    #    17: -
    18: "stat",
    19: "lseek",
    20: "getpid",
    21: "mount",
    22: "oldumount",
    23: "setuid16",
    24: "getuid16",
    25: "stime",
    26: "ptrace",
    27: "alarm",
    28: "fstat",
    29: "pause",
    30: "utime",
    #    31: -
    #    32: -
    33: "access",
    34: "nice",
    #    35: -
    36: "sync",
    37: "kill",
    38: "rename",
    39: "mkdir",
    40: "rmdir",
    41: "dup",
    42: "pipe",
    43: "times",
    #    44: -
    45: "brk",
    46: "setgid16",
    47: "getgid16",
    48: "signal",
    49: "geteuid16",
    50: "getegid16",
    51: "acct",
    52: "umount",
    #    53: -
    54: "ioctl",
    55: "fcntl",
    #    56: -
    57: "setpgid",
    #    58: -
    59: "oldolduname",
    60: "umask",
    61: "chroot",
    62: "ustat",
    63: "dup2",
    64: "getppid",
    65: "getpgrp",
    66: "setsid",
    67: "sigaction",
    68: "sgetmask",
    69: "ssetmask",
    70: "setreuid16",
    71: "setregid16",
    72: "sigsuspend",
    73: "sigpending",
    74: "sethostname",
    75: "setrlimit",
    76: "old_getrlimit",
    77: "getrusage",
    78: "gettimeofday",
    79: "settimeofday",
    80: "getgroups16",
    81: "setgroups16",
    82: "old_select",
    83: "symlink",
    84: "lstat",
    85: "readlink",
    86: "uselib",
    87: "swapon",
    88: "reboot",
    89: "old_readdir",
    90: "old_mmap",
    91: "munmap",
    92: "truncate",
    93: "ftruncate",
    94: "fchmod",
    95: "fchown16",
    96: "getpriority",
    97: "setpriority",
    #    98: -
    99: "statfs",
    100: "fstatfs",
    101: "ioperm",
    102: "socketcall",
    103: "syslog",
    104: "setitimer",
    105: "getitimer",
    106: "newstat",
    107: "newlstat",
    108: "newfstat",
    109: "olduname",
    110: "iopl",
    111: "vhangup",
    #   112: old "idle"
    113: "vm86old",
    114: "wait4",
    115: "swapoff",
    116: "sysinfo",
    117: "ipc",
    118: "fsync",
    119: "sigreturn",
    120: "clone",
    121: "setdomainname",
    122: "uname",
    123: "modify_ldt",
    124: "adjtimex",
    125: "mprotect",
    126: "sigprocmask",
    #   127: old "create_module"
    128: "init_module",
    129: "delete_module",
    #   130: old "get_kernel_syms"
    131: "quotactl",
    132: "getpgid",
    133: "fchdir",
    134: "bdflush",
    135: "sysfs",
    136: "personality",
    #   137: reserved for afs_syscall
    138: "setfsuid16",
    139: "setfsgid16",
    140: "llseek",
    141: "getdents",
    142: "select",
    143: "flock",
    144: "msync",
    145: "readv",
    146: "writev",
    147: "getsid",
    148: "fdatasync",
    149: "sysctl",
    150: "mlock",
    151: "munlock",
    152: "mlockall",
    153: "munlockall",
    154: "sched_setparam",
    155: "sched_getparam",
    156: "sched_setscheduler",
    157: "sched_getscheduler",
    158: "sched_yield",
    159: "sched_get_priority_max",
    160: "sched_get_priority_min",
    161: "sched_rr_get_interval",
    162: "nanosleep",
    163: "mremap",
    164: "setresuid16",
    165: "getresuid16",
    166: "vm86",
    #   167: old "query_module"
    168: "poll",
    169: "nfsservctl",
    170: "setresgid16",
    171: "getresgid16",
    172: "prctl",
    173: "rt_sigreturn",
    174: "rt_sigaction",
    175: "rt_sigprocmask",
    176: "rt_sigpending",
    177: "rt_sigtimedwait",
    178: "rt_sigqueueinfo",
    179: "rt_sigsuspend",
    180: "pread64",
    181: "pwrite64",
    182: "chown16",
    183: "getcwd",
    184: "capget",
    185: "capset",
    186: "sigaltstack",
    187: "sendfile",
    #   188: (reserved)
    #   189: (reserved)
    190: "vfork",
    191: "getrlimit",
    192: "mmap2",
    193: "truncate64",
    194: "ftruncate64",
    195: "stat64",
    196: "lstat64",
    197: "fstat64",
    198: "lchown",
    199: "getuid",
    200: "getgid",
    201: "geteuid",
    202: "getegid",
    203: "setreuid",
    204: "setregid",
    205: "getgroups",
    206: "setgroups",
    207: "fchown",
    208: "setresuid",
    209: "getresuid",
    210: "setresgid",
    211: "getresgid",
    212: "chown",
    213: "setuid",
    214: "setgid",
    215: "setfsuid",
    216: "setfsgid",
    # -------------------------------
    217: "pivot_root",
    218: "mincore",
    219: "madvise",
    220: "getdents64",
    221: "fcntl64",
    #   222: -
    #   223: -
    224: "gettid",
    225: "readahead",
    226: "setxattr",
    227: "lsetxattr",
    228: "fsetxattr",
    229: "getxattr",
    230: "lgetxattr",
    231: "fgetxattr",
    232: "listxattr",
    233: "llistxattr",
    234: "flistxattr",
    235: "removexattr",
    236: "lremovexattr",
    237: "fremovexattr",
    238: "tkill",
    239: "sendfile64",
    240: "futex",
    241: "sched_setaffinity",
    242: "sched_getaffinity",
    243: "set_thread_area",
    244: "get_thread_area",
    245: "io_setup",
    246: "io_destroy",
    247: "io_getevents",
    248: "io_submit",
    249: "io_cancel",
    250: "fadvise64",
    #   251: -
    252: "exit_group",
    253: "lookup_dcookie",
    254: "epoll_create",
    255: "epoll_ctl",
    256: "epoll_wait",
    257: "remap_file_pages",
    258: "set_tid_address",
    259: "timer_create",
    260: "timer_settime",
    261: "timer_gettime",
    262: "timer_getoverrun",
    263: "timer_delete",
    264: "clock_settime",
    265: "clock_gettime",
    266: "clock_getres",
    267: "clock_nanosleep",
    268: "statfs64",
    269: "fstatfs64",
    270: "tgkill",
    271: "utimes",
    272: "fadvise64_64",
    #   273: -
    274: "mbind",
    275: "get_mempolicy",
    276: "set_mempolicy",
    277: "mq_open",
    278: "mq_unlink",
    279: "mq_timedsend",
    280: "mq_timedreceive",
    281: "mq_notify",
    282: "mq_getsetattr",
    283: "kexec_load",
    284: "waitid",
    #   285: -
    286: "add_key",
    287: "request_key",
    288: "keyctl",
    289: "ioprio_set",
    290: "ioprio_get",
    291: "inotify_init",
    292: "inotify_add_watch",
    293: "inotify_rm_watch",
    294: "migrate_pages",
    295: "openat",
    296: "mkdirat",
    297: "mknodat",
    298: "fchownat",
    299: "futimesat",
    300: "fstatat64",
    301: "unlinkat",
    302: "renameat",
    303: "linkat",
    304: "symlinkat",
    305: "readlinkat",
    306: "fchmodat",
    307: "faccessat",
    308: "pselect6",
    309: "ppoll",
    310: "unshare",
    311: "set_robust_list",
    312: "get_robust_list",
    313: "splice",
    314: "sync_file_range",
    315: "tee",
    316: "vmsplice",
    317: "move_pages",
    318: "getcpu",
    319: "epoll_pwait",
}

SOCKET_SYSCALL_NAMES = set(("socketcall",))



from ptrace.os_tools import HAS_PROC
from ptrace.debugger.process_error import ProcessError
from ptrace.ctypes_tools import formatAddress
import re
from weakref import ref

if HAS_PROC:
    from ptrace.linux_proc import openProc, ProcError

PROC_MAP_REGEX = re.compile(
    # Address range: '08048000-080b0000 '
    r'([0-9a-f]+)-([0-9a-f]+) '
    # Permission: 'r-xp '
    r'(.{4}) '
    # Offset: '0804d000'
    r'([0-9a-f]+) '
    # Device (major:minor): 'fe:01 '
    r'([0-9a-f]{2,3}):([0-9a-f]{2}) '
    # Inode: '3334030'
    r'([0-9]+)'
    # Filename: '  /usr/bin/synergyc'
    r'(?: +(.*))?')


class MemoryMapping(object):
    """
    Process memory mapping (metadata about the mapping).

    Attributes:
     - start (int): first byte address
     - end (int): last byte address + 1
     - permissions (str)
     - offset (int): for file, offset in bytes from the file start
     - major_device / minor_device (int): major / minor device number
     - inode (int)
     - pathname (str)
     - _process: weak reference to the process

    Operations:
     - "address in mapping" checks the address is in the mapping.
     - "search(somestring)" returns the offsets of "somestring" in the mapping
     - "str(mapping)" create one string describing the mapping
     - "repr(mapping)" create a string representation of the mapping,
       useful in list contexts
    """

    def __init__(self, process, start, end, permissions, offset, major_device, minor_device, inode, pathname):
        self._process = ref(process)
        self.start = start
        self.end = end
        self.permissions = permissions
        self.offset = offset
        self.major_device = major_device
        self.minor_device = minor_device
        self.inode = inode
        self.pathname = pathname

    def __contains__(self, address):
        return self.start <= address < self.end

    def __str__(self):
        text = "%s-%s" % (formatAddress(self.start), formatAddress(self.end))
        if self.pathname:
            text += " => %s" % self.pathname
        text += " (%s)" % self.permissions
        return text
    __repr__ = __str__

    def search(self, bytestr):
        process = self._process()

        bytestr_len = len(bytestr)
        buf_len = 64 * 1024

        if buf_len < bytestr_len:
            buf_len = bytestr_len

        remaining = self.end - self.start
        covered = self.start

        while remaining >= bytestr_len:
            if remaining > buf_len:
                requested = buf_len
            else:
                requested = remaining

            data = process.readBytes(covered, requested)

            if data == "":
                break

            offset = data.find(bytestr)
            if (offset == -1):
                skip = requested - bytestr_len + 1
            else:
                yield (covered + offset)
                skip = offset + bytestr_len

            covered += skip
            remaining -= skip


def readProcessMappings(process):
    """
    Read all memory mappings of the specified process.

    Return a list of MemoryMapping objects, or empty list if it's not possible
    to read the mappings.

    May raise a ProcessError.
    """
    maps = []
    if not HAS_PROC:
        return maps
    try:
        mapsfile = openProc("%s/maps" % process.pid)
    except ProcError as err:
        raise ProcessError(process, "Unable to read process maps: %s" % err)
    try:
        for line in mapsfile:
            line = line.rstrip()
            match = PROC_MAP_REGEX.match(line)
            if not match:
                raise ProcessError(
                    process, "Unable to parse memory mapping: %r" % line)
            map = MemoryMapping(
                process,
                int(match.group(1), 16),
                int(match.group(2), 16),
                match.group(3),
                int(match.group(4), 16),
                int(match.group(5), 16),
                int(match.group(6), 16),
                int(match.group(7)),
                match.group(8))
            maps.append(map)
    finally:
        mapsfile.close()
    return maps



"""
Constants about the CPU:

 - CPU_BIGENDIAN (bool)
 - CPU_64BITS (bool)
 - CPU_WORD_SIZE (int)
 - CPU_MAX_UINT (int)
 - CPU_PPC32 (bool)
 - CPU_PPC64 (bool)
 - CPU_I386 (bool)
 - CPU_X86_64 (bool)
 - CPU_INTEL (bool)
 - CPU_POWERPC (bool)
"""

try:
    from os import uname
    HAS_UNAME = True
except ImportError:
    HAS_UNAME = False
    from platform import architecture
from sys import byteorder
from ctypes import sizeof, c_void_p

CPU_BIGENDIAN = (byteorder == 'big')
CPU_64BITS = (sizeof(c_void_p) == 8)

if CPU_64BITS:
    CPU_WORD_SIZE = 8  # bytes
    CPU_MAX_UINT = 0xffffffffffffffff
else:
    CPU_WORD_SIZE = 4  # bytes
    CPU_MAX_UINT = 0xffffffff

if HAS_UNAME:
    # guess machine type using uname()
    _machine = uname()[4]
    CPU_PPC32 = (_machine == 'ppc')
    CPU_PPC64 = (_machine in ('ppc64', 'ppc64le'))
    CPU_I386 = (_machine in ("i386", "i686"))    # compatible Intel 32 bits
    CPU_X86_64 = (_machine in ("x86_64", "amd64"))  # compatible Intel 64 bits
    CPU_ARM32 = _machine.startswith('arm')
    CPU_AARCH64 = (_machine == 'aarch64')
    del _machine
else:
    # uname() fallback for Windows
    # I hope that your Windows doesn't run on PPC32/PPC64
    CPU_PPC32 = False
    CPU_PPC64 = False
    CPU_I386 = False
    CPU_X86_64 = False
    CPU_ARM32 = False
    CPU_AARCH64 = False
    bits, linkage = architecture()
    if bits == '32bit':
        CPU_I386 = True
    elif bits == '64bit':
        CPU_X86_64 = True
    else:
        raise ValueError("Unknown architecture bits: %r" % bits)

CPU_INTEL = (CPU_I386 or CPU_X86_64)
CPU_POWERPC = (CPU_PPC32 or CPU_PPC64)
CPU_ARM = (CPU_ARM32 or CPU_AARCH64)



from ptrace.error import PTRACE_ERRORS, writeError
from logging import getLogger
from ptrace.ctypes_tools import formatAddress


class FunctionArgument(object):
    """
    Description of a function argument. Attributes:
     - function: a Function object
     - index (int): index of the argument (starting at zero)
     - options: a FunctionCallOptions object
     - value (int)
     - type (str, optional)
     - text (str): string describing the argument

    Don't use text attribute directly, use getText() to format the
    argument instead.
    """

    def __init__(self, function, index, options,
                 value=None, type=None, name=None):
        self.function = function
        self.index = index
        self.options = options
        self.value = value
        self.type = type
        self.name = name
        self.text = None

    def getText(self):
        if not self.text:
            try:
                text = self.createText()
                if text is not None:
                    self.text = str(text)
                elif self.type and self.type.endswith("*"):
                    self.text = formatAddress(self.value)
                else:
                    self.text = repr(self.value)
            except PTRACE_ERRORS as err:
                writeError(getLogger(), err,
                           "Format argument %s of function %s() value error"
                           % (self.name, self.function.name))
                self.text = repr(self.value)
        return self.text

    def format(self):
        text = self.getText()
        options = self.options
        if options.write_argname and self.name:
            if options.write_types and self.type:
                return "%s %s=%s" % (self.type, self.name, text)
            else:
                return "%s=%s" % (self.name, text)
        elif options.write_types and self.type:
            return "(%s)%s" % (self.type, text)
        else:
            return text

    def createText(self):
        return repr(self.value)

    def formatPointer(self, value, address):
        if self.options.write_address:
            return "%s at %s" % (value, formatAddress(address))
        else:
            return value

    def readStruct(self, address, struct):
        address = self.value

        struct_name = struct.__name__
        data = self.function.process.readStruct(address, struct)
        arguments = []
        for name, argtype in struct._fields_:
            value = getattr(data, name)
            try:
                text = self.formatStructValue(struct_name, name, value)
                if text is not None:
                    text = str(text)
                else:
                    text = repr(value)
            except PTRACE_ERRORS as err:
                writeError(getLogger(), err, "Format struct value error")
                text = repr(value)
            arguments.append("%s=%s" % (name, text))

        data = "<%s %s>" % (struct_name, ", ".join(arguments))
        return self.formatPointer(data, address)

    def formatStructValue(self, struct, name, value):
        return None

    def readArray(self, address, basetype, count):
        array = self.function.process.readArray(address, basetype, count)
        arguments = []
        for index in range(count):
            value = array[index]
            value = str(value)
            arguments.append(value)
        arguments = ", ".join(arguments)
        return self.formatPointer("<(%s)>" % arguments, address)

    def __repr__(self):
        return "argument %s of %s()" % (self.name, self.function.name)



# Linux kernel 5.8.0 on Intel Xeon E5-2670
SYSCALL_NAMES = {
    0: "read",
    1: "write",
    2: "open",
    3: "close",
    4: "stat",
    5: "fstat",
    6: "lstat",
    7: "poll",
    8: "lseek",
    9: "mmap",
    10: "mprotect",
    11: "munmap",
    12: "brk",
    13: "rt_sigaction",
    14: "rt_sigprocmask",
    15: "rt_sigreturn",
    16: "ioctl",
    17: "pread64",
    18: "pwrite64",
    19: "readv",
    20: "writev",
    21: "access",
    22: "pipe",
    23: "select",
    24: "sched_yield",
    25: "mremap",
    26: "msync",
    27: "mincore",
    28: "madvise",
    29: "shmget",
    30: "shmat",
    31: "shmctl",
    32: "dup",
    33: "dup2",
    34: "pause",
    35: "nanosleep",
    36: "getitimer",
    37: "alarm",
    38: "setitimer",
    39: "getpid",
    40: "sendfile",
    41: "socket",
    42: "connect",
    43: "accept",
    44: "sendto",
    45: "recvfrom",
    46: "sendmsg",
    47: "recvmsg",
    48: "shutdown",
    49: "bind",
    50: "listen",
    51: "getsockname",
    52: "getpeername",
    53: "socketpair",
    54: "setsockopt",
    55: "getsockopt",
    56: "clone",
    57: "fork",
    58: "vfork",
    59: "execve",
    60: "exit",
    61: "wait4",
    62: "kill",
    63: "uname",
    64: "semget",
    65: "semop",
    66: "semctl",
    67: "shmdt",
    68: "msgget",
    69: "msgsnd",
    70: "msgrcv",
    71: "msgctl",
    72: "fcntl",
    73: "flock",
    74: "fsync",
    75: "fdatasync",
    76: "truncate",
    77: "ftruncate",
    78: "getdents",
    79: "getcwd",
    80: "chdir",
    81: "fchdir",
    82: "rename",
    83: "mkdir",
    84: "rmdir",
    85: "creat",
    86: "link",
    87: "unlink",
    88: "symlink",
    89: "readlink",
    90: "chmod",
    91: "fchmod",
    92: "chown",
    93: "fchown",
    94: "lchown",
    95: "umask",
    96: "gettimeofday",
    97: "getrlimit",
    98: "getrusage",
    99: "sysinfo",
    100: "times",
    101: "ptrace",
    102: "getuid",
    103: "syslog",
    104: "getgid",
    105: "setuid",
    106: "setgid",
    107: "geteuid",
    108: "getegid",
    109: "setpgid",
    110: "getppid",
    111: "getpgrp",
    112: "setsid",
    113: "setreuid",
    114: "setregid",
    115: "getgroups",
    116: "setgroups",
    117: "setresuid",
    118: "getresuid",
    119: "setresgid",
    120: "getresgid",
    121: "getpgid",
    122: "setfsuid",
    123: "setfsgid",
    124: "getsid",
    125: "capget",
    126: "capset",
    127: "rt_sigpending",
    128: "rt_sigtimedwait",
    129: "rt_sigqueueinfo",
    130: "rt_sigsuspend",
    131: "sigaltstack",
    132: "utime",
    133: "mknod",
    134: "uselib",
    135: "personality",
    136: "ustat",
    137: "statfs",
    138: "fstatfs",
    139: "sysfs",
    140: "getpriority",
    141: "setpriority",
    142: "sched_setparam",
    143: "sched_getparam",
    144: "sched_setscheduler",
    145: "sched_getscheduler",
    146: "sched_get_priority_max",
    147: "sched_get_priority_min",
    148: "sched_rr_get_interval",
    149: "mlock",
    150: "munlock",
    151: "mlockall",
    152: "munlockall",
    153: "vhangup",
    154: "modify_ldt",
    155: "pivot_root",
    156: "_sysctl",
    157: "prctl",
    158: "arch_prctl",
    159: "adjtimex",
    160: "setrlimit",
    161: "chroot",
    162: "sync",
    163: "acct",
    164: "settimeofday",
    165: "mount",
    166: "umount2",
    167: "swapon",
    168: "swapoff",
    169: "reboot",
    170: "sethostname",
    171: "setdomainname",
    172: "iopl",
    173: "ioperm",
    174: "create_module",
    175: "init_module",
    176: "delete_module",
    177: "get_kernel_syms",
    178: "query_module",
    179: "quotactl",
    180: "nfsservctl",
    181: "getpmsg",
    182: "putpmsg",
    183: "afs_syscall",
    184: "tuxcall",
    185: "security",
    186: "gettid",
    187: "readahead",
    188: "setxattr",
    189: "lsetxattr",
    190: "fsetxattr",
    191: "getxattr",
    192: "lgetxattr",
    193: "fgetxattr",
    194: "listxattr",
    195: "llistxattr",
    196: "flistxattr",
    197: "removexattr",
    198: "lremovexattr",
    199: "fremovexattr",
    200: "tkill",
    201: "time",
    202: "futex",
    203: "sched_setaffinity",
    204: "sched_getaffinity",
    205: "set_thread_area",
    206: "io_setup",
    207: "io_destroy",
    208: "io_getevents",
    209: "io_submit",
    210: "io_cancel",
    211: "get_thread_area",
    212: "lookup_dcookie",
    213: "epoll_create",
    214: "epoll_ctl_old",
    215: "epoll_wait_old",
    216: "remap_file_pages",
    217: "getdents64",
    218: "set_tid_address",
    219: "restart_syscall",
    220: "semtimedop",
    221: "fadvise64",
    222: "timer_create",
    223: "timer_settime",
    224: "timer_gettime",
    225: "timer_getoverrun",
    226: "timer_delete",
    227: "clock_settime",
    228: "clock_gettime",
    229: "clock_getres",
    230: "clock_nanosleep",
    231: "exit_group",
    232: "epoll_wait",
    233: "epoll_ctl",
    234: "tgkill",
    235: "utimes",
    236: "vserver",
    237: "mbind",
    238: "set_mempolicy",
    239: "get_mempolicy",
    240: "mq_open",
    241: "mq_unlink",
    242: "mq_timedsend",
    243: "mq_timedreceive",
    244: "mq_notify",
    245: "mq_getsetattr",
    246: "kexec_load",
    247: "waitid",
    248: "add_key",
    249: "request_key",
    250: "keyctl",
    251: "ioprio_set",
    252: "ioprio_get",
    253: "inotify_init",
    254: "inotify_add_watch",
    255: "inotify_rm_watch",
    256: "migrate_pages",
    257: "openat",
    258: "mkdirat",
    259: "mknodat",
    260: "fchownat",
    261: "futimesat",
    262: "newfstatat",
    263: "unlinkat",
    264: "renameat",
    265: "linkat",
    266: "symlinkat",
    267: "readlinkat",
    268: "fchmodat",
    269: "faccessat",
    270: "pselect6",
    271: "ppoll",
    272: "unshare",
    273: "set_robust_list",
    274: "get_robust_list",
    275: "splice",
    276: "tee",
    277: "sync_file_range",
    278: "vmsplice",
    279: "move_pages",
    280: "utimensat",
    281: "epoll_pwait",
    282: "signalfd",
    283: "timerfd",
    284: "eventfd",
    285: "fallocate",
    286: "timerfd_settime",
    287: "timerfd_gettime",
    288: "accept4",
    289: "signalfd4",
    290: "eventfd2",
    291: "epoll_create1",
    292: "dup3",
    293: "pipe2",
    294: "inotify_init1",
    295: "preadv",
    296: "pwritev",
    297: "rt_tgsigqueueinfo",
    298: "perf_event_open",
    299: "recvmmsg",
    300: "fanotify_init",
    301: "fanotify_mark",
    302: "prlimit64",
    303: "name_to_handle_at",
    304: "open_by_handle_at",
    305: "clock_adjtime",
    306: "syncfs",
    307: "sendmmsg",
    308: "setns",
    309: "getcpu",
    310: "process_vm_readv",
    311: "process_vm_writev",
    312: "kcmp",
    313: "finit_module",
    314: "sched_setattr",
    315: "sched_getattr",
    316: "renameat2",
    317: "seccomp",
    318: "getrandom",
    319: "memfd_create",
    320: "kexec_file_load",
    321: "bpf",
    322: "execveat",
    323: "userfaultfd",
    324: "membarrier",
    325: "mlock2",
    326: "copy_file_range",
    327: "preadv2",
    328: "pwritev2",
    329: "pkey_mprotect",
    330: "pkey_alloc",
    331: "pkey_free",
    332: "statx",
    333: "io_pgetevents",
    334: "rseq",
    424: "pidfd_send_signal",
    425: "io_uring_setup",
    426: "io_uring_enter",
    427: "io_uring_register",
    428: "open_tree",
    429: "move_mount",
    430: "fsopen",
    431: "fsconfig",
    432: "fsmount",
    433: "fspick",
    434: "pidfd_open",
    435: "clone3",
    436: "close_range",
    437: "openat2",
    438: "pidfd_getfd",
    439: "faccessat2",
    440: "process_madvise",
    512: "rt_sigaction",
    513: "rt_sigreturn",
    514: "ioctl",
    515: "readv",
    516: "writev",
    517: "recvfrom",
    518: "sendmsg",
    519: "recvmsg",
    520: "execve",
    521: "ptrace",
    522: "rt_sigpending",
    523: "rt_sigtimedwait",
    524: "rt_sigqueueinfo",
    525: "sigaltstack",
    526: "timer_create",
    527: "mq_notify",
    528: "kexec_load",
    529: "waitid",
    530: "set_robust_list",
    531: "get_robust_list",
    532: "vmsplice",
    533: "move_pages",
    534: "preadv",
    535: "pwritev",
    536: "rt_tgsigqueueinfo",
    537: "recvmmsg",
    538: "sendmmsg",
    539: "process_vm_readv",
    540: "process_vm_writev",
    541: "setsockopt",
    542: "getsockopt",
    543: "io_setup",
    544: "io_submit",
    545: "execveat",
    546: "preadv2",
    547: "pwritev2",
}

SOCKET_SYSCALL_NAMES = set((
    "socket", "socketpair", "connect",
    "sendto", "recvfrom", "sendmsg", "recvmsg",
    "bind", "listen", "accept",
    "getsockname", "getpeername", "getsockopt", "setsockopt",
    "shutdown",
))



from logging import info, warning, error
from ptrace import PtraceError
from os import waitpid, WNOHANG
from signal import SIGTRAP, SIGSTOP
from errno import ECHILD
from ptrace.debugger import PtraceProcess, ProcessSignal
from ptrace.binding import HAS_PTRACE_EVENTS
from time import sleep
if HAS_PTRACE_EVENTS:
    from ptrace.binding.func import (
        PTRACE_O_TRACEFORK, PTRACE_O_TRACEVFORK,
        PTRACE_O_TRACEEXEC, PTRACE_O_TRACESYSGOOD,
        PTRACE_O_TRACECLONE, THREAD_TRACE_FLAGS)


class DebuggerError(PtraceError):
    pass


class PtraceDebugger(object):
    """
    Debugger managing one or multiple processes at the same time.

    Methods
    =======

     * Process list:
       - addProcess(): add a new process
       - deleteProcess(): remove a process from the debugger

     * Wait for an event:
       - waitProcessEvent(): wait for a process event
       - waitSignals(): wait for a signal
       - waitSyscall(): wait for the next syscall event

     * Options:
      - traceFork(): enable fork() tracing
      - traceExec(): enable exec() tracing
      - traceClone(): enable clone() tracing
      - enableSysgood(): enable sysgood option

     * Other:
       - quit(): quit the debugger, terminate all processes

    Operations
    ==========

     - iterate on all processes: "for process in debugger: ..."
     - get a process by its identifier: "process = debugger[pid]"
     - get the number of processes: len(debugger)

    Attributes
    ==========

     - dict: processes dictionary (pid -> PtraceProcess)
     - list: processes list
     - options: ptrace options
     - trace_fork (bool): fork() tracing is enabled?
     - trace_exec (bool): exec() tracing is enabled?
     - trace_clone (bool): clone() tracing is enabled?
     - use_sysgood (bool): sysgood option is enabled?
    """

    def __init__(self):
        self.dict = {}   # pid -> PtraceProcess object
        self.list = []
        self.options = 0
        self.trace_fork = False
        self.trace_exec = False
        self.trace_clone = False
        self.use_sysgood = False
        self.enableSysgood()

    def addProcess(self, pid, is_attached, parent=None, is_thread=False):
        """
        Add a new process using its identifier. Use is_attached=False to
        attach an existing (running) process, and is_attached=True to trace
        a new (stopped) process.
        """
        if pid in self.dict:
            raise KeyError("The process %s is already registered!" % pid)
        process = PtraceProcess(self, pid, is_attached,
                                parent=parent, is_thread=is_thread)
        info("Attach %s to debugger" % process)
        self.dict[pid] = process
        self.list.append(process)
        try:
            process.waitSignals(SIGTRAP, SIGSTOP)
        except KeyboardInterrupt:
            error(
                "User interrupt! Force the process %s attach "
                "(don't wait for signals)."
                % pid)
        except ProcessSignal as event:
            event.display()
        except:   # noqa: E722
            process.is_attached = False
            process.detach()
            raise
        if HAS_PTRACE_EVENTS and self.options:
            process.setoptions(self.options)
        return process

    def quit(self):
        """
        Quit the debugger: terminate all processes in reverse order.
        """
        info("Quit debugger")
        # Terminate processes in reverse order
        # to kill children before parents
        processes = list(self.list)
        for process in reversed(processes):
            process.terminate()
            process.detach()

    def _waitpid(self, wanted_pid, blocking=True):
        """
        Wait for a process event from a specific process (if wanted_pid is
        set) or any process (wanted_pid=None). The call is blocking is
        blocking option is True. Return the tuple (pid, status).

        See os.waitpid() documentation for explanations about the result.
        """
        flags = 0
        if not blocking:
            flags |= WNOHANG
        if wanted_pid:
            if wanted_pid not in self.dict:
                raise DebuggerError("Unknown PID: %r" %
                                    wanted_pid, pid=wanted_pid)

            process = self.dict[wanted_pid]
            if process.is_thread:
                flags |= THREAD_TRACE_FLAGS

            pid, status = waitpid(wanted_pid, flags)
        else:
            pid, status = waitpid(-1, flags)
        if (blocking or pid) and wanted_pid and (pid != wanted_pid):
            raise DebuggerError("Unwanted PID: %r (instead of %s)"
                                % (pid, wanted_pid), pid=pid)
        return pid, status

    def _wait_event_pid(self, wanted_pid, blocking=True):
        """
        Wait for a process event from the specified process identifier. If
        blocking=False, return None if there is no new event, otherwise return
        an object based on ProcessEvent.
        """
        process = None
        while not process:
            try:
                pid, status = self._waitpid(wanted_pid, blocking)
            except OSError as err:
                if err.errno == ECHILD:
                    process = self.dict[wanted_pid]
                    return process.processTerminated()
                else:
                    raise err
            if not blocking and not pid:
                return None
            try:
                process = self.dict[pid]
            except KeyError:
                warning("waitpid() warning: Unknown PID %r" % pid)
        return process.processStatus(status)

    def _wait_event(self, wanted_pid, blocking=True):
        if wanted_pid is not None:
            return self._wait_event_pid(wanted_pid, blocking)

        pause = 0.001
        while True:
            pids = tuple(self.dict)
            if len(pids) > 1:
                for pid in pids:
                    process = self._wait_event_pid(pid, False)
                    if process is not None:
                        return process
                if not blocking:
                    return None
                pause = min(pause * 2, 0.5)
                sleep(pause)
            else:
                return self._wait_event_pid(pids[0], blocking)

    def waitProcessEvent(self, pid=None, blocking=True):
        """
        Wait for a process event from a specific process (if pid option is
        set) or any process (default). If blocking=False, return None if there
        is no new event, otherwise return an object based on ProcessEvent.
        """
        return self._wait_event(pid, blocking)

    def waitSignals(self, *signals, **kw):
        """
        Wait for any signal or some specific signals (if specified) from a
        specific process (if pid keyword is set) or any process (default).
        Return a ProcessSignal object or raise an unexpected ProcessEvent.
        """
        pid = kw.get('pid', None)
        while True:
            event = self._wait_event(pid)
            if event.__class__ != ProcessSignal:
                raise event
            signum = event.signum
            if signum in signals or not signals:
                return event
            raise event

    def waitSyscall(self, process=None):
        """
        Wait for the next syscall event (enter or exit) for a specific process
        (if specified) or any process (default). Return a ProcessSignal object
        or raise an unexpected ProcessEvent.
        """
        signum = SIGTRAP
        if self.use_sysgood:
            signum |= 0x80
        if process:
            return self.waitSignals(signum, pid=process.pid)
        else:
            return self.waitSignals(signum)

    def deleteProcess(self, process=None, pid=None):
        """
        Delete a process from the process list.
        """
        if not process:
            try:
                process = self.dict[pid]
            except KeyError:
                return
        try:
            del self.dict[process.pid]
        except KeyError:
            pass
        try:
            self.list.remove(process)
        except ValueError:
            pass

    def traceFork(self):
        """
        Enable fork() tracing. Do nothing if it's not supported.
        """
        if not HAS_PTRACE_EVENTS:
            raise DebuggerError(
                "Tracing fork events is not supported on this architecture or operating system")
        self.options |= PTRACE_O_TRACEFORK | PTRACE_O_TRACEVFORK
        self.trace_fork = True
        info("Debugger trace forks (options=%s)" % self.options)

    def traceExec(self):
        """
        Enable exec() tracing. Do nothing if it's not supported.
        """
        if not HAS_PTRACE_EVENTS:
            # no effect on OS without ptrace events
            return
        self.trace_exec = True
        self.options |= PTRACE_O_TRACEEXEC

    def traceClone(self):
        """
        Enable clone() tracing. Do nothing if it's not supported.
        """
        if not HAS_PTRACE_EVENTS:
            # no effect on OS without ptrace events
            return
        self.trace_clone = True
        self.options |= PTRACE_O_TRACECLONE

    def enableSysgood(self):
        """
        Enable sysgood option: ask the kernel to set bit #7 of the signal
        number if the signal comes from the kernel space. If the signal comes
        from the user space, the bit is unset.
        """
        if not HAS_PTRACE_EVENTS:
            # no effect on OS without ptrace events
            return
        self.use_sysgood = True
        self.options |= PTRACE_O_TRACESYSGOOD

    def __getitem__(self, pid):
        return self.dict[pid]

    def __iter__(self):
        return iter(self.list)

    def __len__(self):
        return len(self.list)



from ptrace.cpu_info import (
    CPU_POWERPC, CPU_INTEL, CPU_X86_64, CPU_I386, CPU_ARM32, CPU_AARCH64)

CPU_INSTR_POINTER = None
CPU_STACK_POINTER = None
CPU_FRAME_POINTER = None
CPU_SUB_REGISTERS = {}

if CPU_POWERPC:
    CPU_INSTR_POINTER = "nip"
    # FIXME: Is it the right register?
    CPU_STACK_POINTER = 'gpr1'
elif CPU_ARM32:
    CPU_INSTR_POINTER = 'r15'
    CPU_STACK_POINTER = 'r14'
    CPU_FRAME_POINTER = 'r11'
elif CPU_AARCH64:
    CPU_INSTR_POINTER = 'pc'
    CPU_STACK_POINTER = 'sp'
    CPU_FRAME_POINTER = 'r29'
elif CPU_X86_64:
    CPU_INSTR_POINTER = "rip"
    CPU_STACK_POINTER = "rsp"
    CPU_FRAME_POINTER = "rbp"
    CPU_SUB_REGISTERS = {
        # main register name, shift, mask
        'al': ('rax', 0, 0xff),
        'bl': ('rbx', 0, 0xff),
        'cl': ('rcx', 0, 0xff),
        'dl': ('rdx', 0, 0xff),
        'ah': ('rax', 8, 0xff),
        'bh': ('rbx', 8, 0xff),
        'ch': ('rcx', 8, 0xff),
        'dh': ('rdx', 8, 0xff),
        'ax': ('rax', 0, 0xffff),
        'bx': ('rbx', 0, 0xffff),
        'cx': ('rcx', 0, 0xffff),
        'dx': ('rdx', 0, 0xffff),
        'eax': ('rax', 32, None),
        'ebx': ('rbx', 32, None),
        'ecx': ('rcx', 32, None),
        'edx': ('rdx', 32, None),
    }
elif CPU_I386:
    CPU_INSTR_POINTER = "eip"
    CPU_STACK_POINTER = "esp"
    CPU_FRAME_POINTER = "ebp"
    CPU_SUB_REGISTERS = {
        'al': ('eax', 0, 0xff),
        'bl': ('ebx', 0, 0xff),
        'cl': ('ecx', 0, 0xff),
        'dl': ('edx', 0, 0xff),
        'ah': ('eax', 8, 0xff),
        'bh': ('ebx', 8, 0xff),
        'ch': ('ecx', 8, 0xff),
        'dh': ('edx', 8, 0xff),
        'ax': ('eax', 0, 0xffff),
        'bx': ('ebx', 0, 0xffff),
        'cx': ('ecx', 0, 0xffff),
        'dx': ('edx', 0, 0xffff),
    }

if CPU_INTEL:
    CPU_SUB_REGISTERS.update({
        'cf': ('eflags', 0, 1),
        'pf': ('eflags', 2, 1),
        'af': ('eflags', 4, 1),
        'zf': ('eflags', 6, 1),
        'sf': ('eflags', 7, 1),
        'tf': ('eflags', 8, 1),
        'if': ('eflags', 9, 1),
        'df': ('eflags', 10, 1),
        'of': ('eflags', 11, 1),
        'iopl': ('eflags', 12, 2),
    })
